"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.queryFromArchivers = exports.RequestDataType = exports.validateRequestData = exports.registerRoutes = void 0;
const Config_1 = require("./Config");
const Crypto = require("./Crypto");
const State = require("./State");
const NodeList = require("./NodeList");
const P2P = require("./P2P");
const Storage = require("./archivedCycle/Storage");
const Data = require("./Data/Data");
const Cycles = require("./Data/Cycles");
const Utils = require("./Utils");
const Gossip_1 = require("./archivedCycle/Gossip");
const Logger = require("./Logger");
const lib_types_1 = require("@shardeum-foundation/lib-types");
const stream_1 = require("stream");
const nestedCounters_1 = require("./profiler/nestedCounters");
const profiler_1 = require("./profiler/profiler");
const CycleDB = require("./dbstore/cycles");
const AccountDB = require("./dbstore/accounts");
const TransactionDB = require("./dbstore/transactions");
const ReceiptDB = require("./dbstore/receipts");
const OriginalTxDB = require("./dbstore/originalTxsData");
const Collector = require("./Data/Collector");
const AccountDataProvider = require("./Data/AccountDataProvider");
const GlobalAccount_1 = require("./GlobalAccount");
const Cycles_1 = require("./Data/Cycles");
const DebugMode_1 = require("./DebugMode");
const lib_types_2 = require("@shardeum-foundation/lib-types");
const primary_process_1 = require("./primary-process");
const ServiceQueue = require("./ServiceQueue");
const tickets_1 = require("./routes/tickets");
const allowedArchiversManager_1 = require("./shardeum/allowedArchiversManager");
const CheckpointData_1 = require("./checkpoint/CheckpointData");
const Utils_1 = require("./checkpoint/Utils");
const checkpointStatus_1 = require("./dbstore/checkpointStatus");
const archiverLogging_1 = require("./profiler/archiverLogging");
const { version } = require('../package.json'); // eslint-disable-line @typescript-eslint/no-var-requires
const TXID_LENGTH = 64;
const { MAX_CYCLES_PER_REQUEST, MAX_ORIGINAL_TXS_PER_REQUEST, MAX_RECEIPTS_PER_REQUEST, MAX_ACCOUNTS_PER_REQUEST, MAX_BETWEEN_CYCLES_PER_REQUEST, } = Config_1.config.REQUEST_LIMIT;
let reachabilityAllowed = true;
function registerRoutes(server) {
    server.get('/myip', function (request, reply) {
        const ip = request.raw.socket.remoteAddress;
        reply.send({ ip });
    });
    server.post('/nodelist', (request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('POST_nodelist');
        try {
            nestedCounters_1.nestedCountersInstance.countEvent('consensor', 'POST_nodelist', 1);
            const signedFirstNodeInfo = request.body;
            archiverLogging_1.ArchiverLogging.logValidatorConnection({
                validatorId: signedFirstNodeInfo.nodeInfo.publicKey,
                archiverId: Config_1.config.ARCHIVER_IP,
                timestamp: Date.now(),
                status: 'CONNECTING',
                handshake: {
                    success: false,
                    duration: 0,
                },
            });
            if (State.isFirst && NodeList.isEmpty() && !NodeList.foundFirstNode) {
                try {
                    let err = Utils.validateTypes(signedFirstNodeInfo, {
                        nodeInfo: 'o',
                        sign: 'o',
                    });
                    if (err) {
                        archiverLogging_1.ArchiverLogging.logValidatorConnection({
                            validatorId: signedFirstNodeInfo.nodeInfo.publicKey,
                            archiverId: Config_1.config.ARCHIVER_IP,
                            timestamp: Date.now(),
                            status: 'ERROR',
                            handshake: {
                                success: false,
                                duration: Date.now() - request.raw.socket.remotePort,
                                error: err,
                            },
                        });
                        reply.send({ success: false, error: err });
                        return;
                    }
                    err = Utils.validateTypes(signedFirstNodeInfo.nodeInfo, {
                        externalIp: 's',
                        externalPort: 'n',
                        publicKey: 's',
                    });
                    if (err) {
                        reply.send({ success: false, error: err });
                        return;
                    }
                    err = Utils.validateTypes(signedFirstNodeInfo.sign, {
                        owner: 's',
                        sig: 's',
                    });
                    if (err) {
                        reply.send({ success: false, error: err });
                        return;
                    }
                    if (signedFirstNodeInfo.nodeInfo.publicKey !== signedFirstNodeInfo.sign.owner) {
                        Logger.mainLogger.error('nodeInfo.publicKey does not match signature owner', signedFirstNodeInfo);
                        reply.send({ success: false, error: 'nodeInfo.publicKey does not match signature owner' });
                        return;
                    }
                    const isSignatureValid = Crypto.verify(signedFirstNodeInfo);
                    if (!isSignatureValid) {
                        Logger.mainLogger.error('Invalid signature', signedFirstNodeInfo);
                        reply.send({ success: false, error: 'Invalid signature' });
                        return;
                    }
                }
                catch (e) {
                    Logger.mainLogger.error(e);
                    reply.send({ success: false, error: 'Signature verification failed' });
                    return;
                }
                const ip = signedFirstNodeInfo.nodeInfo.externalIp;
                const port = signedFirstNodeInfo.nodeInfo.externalPort;
                const publicKey = signedFirstNodeInfo.nodeInfo.publicKey;
                if (Config_1.config.restrictFirstNodeSelectionByPublicKey) {
                    if (publicKey !== Config_1.config.firstNodePublicKey) {
                        Logger.mainLogger.error('Invalid publicKey of first node info', signedFirstNodeInfo);
                        reply.send({ success: false, error: 'Invalid publicKey of first node info' });
                        return;
                    }
                }
                if (NodeList.foundFirstNode) {
                    const res = NodeList.getCachedNodeList();
                    reply.send(res);
                    return;
                }
                NodeList.toggleFirstNode();
                const firstNode = {
                    ip,
                    port,
                    publicKey,
                };
                Data.initSocketClient(firstNode);
                archiverLogging_1.ArchiverLogging.logValidatorConnection({
                    validatorId: signedFirstNodeInfo.nodeInfo.publicKey,
                    archiverId: Config_1.config.ARCHIVER_IP,
                    timestamp: Date.now(),
                    status: 'CONNECTED',
                    handshake: {
                        success: true,
                        duration: Date.now() - request.raw.socket.remotePort,
                    },
                });
                // Add first node to NodeList
                NodeList.addNodes(NodeList.NodeStatus.SYNCING, [firstNode]);
                // Setting current time for realUpdatedTimes to refresh the nodelist and full-nodelist cache
                NodeList.realUpdatedTimes.set('/nodelist', Date.now());
                NodeList.realUpdatedTimes.set('/full-nodelist', Date.now());
                // Set first node as dataSender
                const firstDataSender = {
                    nodeInfo: firstNode,
                    types: [lib_types_1.P2P.SnapshotTypes.TypeNames.CYCLE, lib_types_1.P2P.SnapshotTypes.TypeNames.STATE_METADATA],
                    contactTimeout: Data.createContactTimeout(firstNode.publicKey, 'This timeout is created for the first node'),
                };
                Data.addDataSender(firstDataSender);
                let res;
                if (Config_1.config.experimentalSnapshot) {
                    const data = {
                        nodeList: [firstNode],
                    };
                    if (Cycles_1.cycleRecordWithShutDownMode) {
                        // For restore network to start the network from the 'restart' mode
                        data['restartCycleRecord'] = Cycles_1.cycleRecordWithShutDownMode;
                        data['dataRequestCycle'] = Cycles_1.cycleRecordWithShutDownMode.counter;
                    }
                    else {
                        // For new network to start the network from the 'forming' mode
                        data['joinRequest'] = P2P.createArchiverJoinRequest();
                        data['dataRequestCycle'] = Cycles.getCurrentCycleCounter();
                    }
                    res = Crypto.sign(data);
                }
                else {
                    res = Crypto.sign({
                        nodeList: [firstNode],
                        joinRequest: P2P.createArchiverJoinRequest(),
                        dataRequestCycle: Data.createDataRequest(lib_types_1.P2P.SnapshotTypes.TypeNames.CYCLE, Cycles.getCurrentCycleCounter(), publicKey),
                        dataRequestStateMetaData: Data.createDataRequest(lib_types_1.P2P.SnapshotTypes.TypeNames.STATE_METADATA, Cycles.lastProcessedMetaData, publicKey),
                    });
                }
                reply.send(res);
            }
            else {
                // Note, this is doing the same thing as GET /nodelist. However, it has been kept for backwards
                // compatibility.
                const res = NodeList.getCachedNodeList();
                reply.send(res);
            }
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('POST_nodelist');
        }
    });
    server.get('/nodelist', (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('GET_nodelist');
        try {
            nestedCounters_1.nestedCountersInstance.countEvent('consensor', 'GET_nodelist');
            const res = NodeList.getCachedNodeList();
            reply.send(res);
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('GET_nodelist');
        }
    });
    server.get('/network-txs-list', (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('GET_network-txs-list');
        try {
            nestedCounters_1.nestedCountersInstance.countEvent('consensor', 'network-txs-list');
            const res = ServiceQueue.getTxList();
            reply.send(res);
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('GET_network-txs-list');
        }
    });
    server.get('/full-nodelist', (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('FULL_nodelist');
        try {
            nestedCounters_1.nestedCountersInstance.countEvent('consensor', 'FULL_nodelist');
            const query = _request.query;
            let activeOnly = false;
            let syncingOnly = false;
            let standbyOnly = false;
            if (query.activeOnly === 'true')
                activeOnly = true;
            if (query.syncingOnly === 'true')
                syncingOnly = true;
            if (query.standbyOnly === 'true')
                standbyOnly = true;
            const res = NodeList.getCachedFullNodeList(activeOnly, syncingOnly, standbyOnly);
            reply.send(res);
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('FULL_nodelist');
        }
    });
    server.get('/removed', {
        preHandler: async (_request, reply) => {
            (0, DebugMode_1.isDebugMiddleware)(_request, reply);
        },
    }, (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('removed');
        nestedCounters_1.nestedCountersInstance.countEvent('consensor', 'removed');
        reply.send({ removedAndApopedNodes: Cycles.removedAndApopedNodes });
        profiler_1.profilerInstance.profileSectionEnd('removed');
    });
    server.get('/archivers', (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('GET_archivers');
        try {
            nestedCounters_1.nestedCountersInstance.countEvent('consensor', 'GET_archivers');
            const activeArchivers = State.activeArchivers
                .filter((archiver) => State.archiversReputation.has(archiver.publicKey) &&
                State.archiversReputation.get(archiver.publicKey) === 'up')
                .map(({ publicKey, ip, port }) => ({ publicKey, ip, port }));
            const res = Crypto.sign({
                activeArchivers,
            });
            reply.send(res);
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('GET_archivers');
        }
    });
    server.get('/allowed-archivers', async (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('GET_allowed_archivers');
        try {
            const config = allowedArchiversManager_1.allowedArchiversManager.getCurrentConfig();
            if (!config) {
                return reply.status(500).send({
                    error: 'Internal server error',
                });
            }
            return reply.send(config);
        }
        catch (error) {
            Logger.mainLogger.error('Error serving allowed-archivers:', error);
            return reply.status(500).send({
                error: 'Internal server error',
            });
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('GET_allowed_archivers');
        }
    });
    server.get('/nodeInfo', (request, reply) => {
        if (reachabilityAllowed) {
            reply.send({
                publicKey: Config_1.config.ARCHIVER_PUBLIC_KEY,
                ip: Config_1.config.ARCHIVER_IP,
                port: Config_1.config.ARCHIVER_PORT,
                version,
                time: Date.now(),
            });
        }
        else {
            request.raw.socket.destroy();
        }
    });
    server.post('/cycleinfo', async (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('POST_cycleinfo');
        try {
            const requestData = _request.body;
            const result = (0, exports.validateRequestData)(requestData, {
                start: 'n?',
                end: 'n?',
                count: 'n?',
                download: 'b?',
                sender: 's',
                sign: 'o',
            }, true);
            if (!result.success) {
                reply.send(Crypto.sign({ success: false, error: result.error }));
                return;
            }
            const { start, end, count, download } = _request.body;
            if (download !== undefined && typeof download !== 'boolean') {
                reply.send(Crypto.sign({ success: false, error: `Invalid download flag` }));
                return;
            }
            const isDownload = download === true;
            let cycleInfo = [];
            if (count) {
                if (count <= 0 || Number.isNaN(count)) {
                    reply.send(Crypto.sign({ success: false, error: `Invalid count` }));
                    return;
                }
                if (count > MAX_CYCLES_PER_REQUEST) {
                    reply.send(Crypto.sign({ success: false, error: `Max count is ${MAX_CYCLES_PER_REQUEST}.` }));
                    return;
                }
                cycleInfo = await CycleDB.queryLatestCycleRecords(count);
            }
            else if (start || start === 0) {
                const from = start;
                const to = end ? end : from;
                if (!(from >= 0 && to >= from) || Number.isNaN(from) || Number.isNaN(to)) {
                    Logger.mainLogger.error(`Invalid start and end counters`);
                    reply.send(Crypto.sign({ success: false, error: `Invalid start and end counters` }));
                    return;
                }
                const cycleCount = to - from;
                if (cycleCount > MAX_CYCLES_PER_REQUEST) {
                    Logger.mainLogger.error(`Exceed maximum limit of ${MAX_CYCLES_PER_REQUEST} cycles`);
                    reply.send(Crypto.sign({ success: false, error: `Exceed maximum limit of ${MAX_CYCLES_PER_REQUEST} cycles` }));
                    return;
                }
                cycleInfo = await CycleDB.queryCycleRecordsBetween(from, to);
                if (isDownload) {
                    const dataInBuffer = Buffer.from(lib_types_2.Utils.safeStringify(cycleInfo), 'utf-8');
                    const dataInStream = stream_1.Readable.from(dataInBuffer);
                    const filename = `cycle_records_from_${from}_to_${to}`;
                    reply.headers({
                        'content-disposition': `attachment; filename="${filename}"`,
                        'content-type': 'application/octet-stream',
                    });
                    reply.send(dataInStream);
                    return;
                }
            }
            else {
                reply.send(Crypto.sign({
                    success: false,
                    error: 'not specified which cycle to show',
                }));
                return;
            }
            const res = Crypto.sign({
                cycleInfo,
            });
            reply.send(res);
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('POST_cycleinfo');
        }
    });
    server.get('/cycleinfo/:count', async (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('GET_cycleinfo');
        try {
            const err = Utils.validateTypes(_request.params, { count: 's' });
            if (err) {
                reply.send({ success: false, error: err });
                return;
            }
            let count = parseInt(_request.params.count);
            if (count <= 0 || Number.isNaN(count)) {
                reply.send({ success: false, error: `Invalid count` });
                return;
            }
            if (count > MAX_CYCLES_PER_REQUEST)
                count = MAX_CYCLES_PER_REQUEST;
            const res = await Cycles.getLatestCycleRecords(count);
            reply.send(res);
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('GET_cycleinfo');
        }
    });
    server.post('/originalTx', async (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('POST_originalTx');
        try {
            const requestData = _request.body;
            const result = (0, exports.validateRequestData)(requestData, {
                count: 'n?',
                start: 'n?',
                end: 'n?',
                startCycle: 'n?',
                endCycle: 'n?',
                type: 's?',
                page: 'n?',
                txId: 's?',
                txIdList: 'a?',
                sender: 's',
                sign: 'o',
            });
            if (!result.success) {
                reply.send(Crypto.sign({ success: false, error: result.error }));
                return;
            }
            const { count, start, end, startCycle, endCycle, type, page, txId, txIdList } = _request.body;
            let originalTxs = [];
            if (count) {
                if (count <= 0 || Number.isNaN(count)) {
                    reply.send(Crypto.sign({ success: false, error: `Invalid count` }));
                    return;
                }
                if (count > MAX_ORIGINAL_TXS_PER_REQUEST) {
                    reply.send(Crypto.sign({ success: false, error: `Max count is ${MAX_ORIGINAL_TXS_PER_REQUEST}` }));
                    return;
                }
                originalTxs = await OriginalTxDB.queryLatestOriginalTxs(count);
            }
            else if (txId) {
                if (txId.length !== TXID_LENGTH) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Invalid txId ${txId}`,
                    }));
                    return;
                }
                const originalTx = await OriginalTxDB.queryOriginalTxDataByTxId(txId);
                if (originalTx)
                    originalTxs.push(originalTx);
            }
            else if (txIdList) {
                if (txIdList.length > MAX_ORIGINAL_TXS_PER_REQUEST) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Exceed maximum limit of ${MAX_ORIGINAL_TXS_PER_REQUEST} original transactions`,
                    }));
                    return;
                }
                for (const [txId, txTimestamp] of txIdList) {
                    if (typeof txId !== 'string' || txId.length !== TXID_LENGTH || typeof txTimestamp !== 'number') {
                        reply.send(Crypto.sign({
                            success: false,
                            error: `Invalid txId ${txId} in the List`,
                        }));
                        return;
                    }
                    const originalTx = await OriginalTxDB.queryOriginalTxDataByTxId(txId, txTimestamp);
                    if (originalTx)
                        originalTxs.push(originalTx);
                }
            }
            else if (start || start === 0) {
                const from = start;
                const to = end ? end : from;
                if (!(from >= 0 && to >= from) || Number.isNaN(from) || Number.isNaN(to)) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Invalid start and end counters`,
                    }));
                    return;
                }
                const count = to - from;
                if (count > MAX_ORIGINAL_TXS_PER_REQUEST) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Exceed maximum limit of ${MAX_ORIGINAL_TXS_PER_REQUEST} original transactions`,
                    }));
                    return;
                }
                originalTxs = await OriginalTxDB.queryOriginalTxsData(from, count + 1);
            }
            else if (startCycle || startCycle === 0) {
                const from = startCycle;
                const to = endCycle ? endCycle : from;
                if (!(from >= 0 && to >= from) || Number.isNaN(from) || Number.isNaN(to)) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Invalid startCycle and endCycle counters`,
                    }));
                    return;
                }
                const count = to - from;
                if (count > MAX_BETWEEN_CYCLES_PER_REQUEST) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Exceed maximum limit of ${MAX_BETWEEN_CYCLES_PER_REQUEST} cycles`,
                    }));
                    return;
                }
                if (type === 'tally') {
                    originalTxs = await OriginalTxDB.queryOriginalTxDataCountByCycles(from, to);
                }
                else if (type === 'count') {
                    originalTxs = await OriginalTxDB.queryOriginalTxDataCount(from, to);
                }
                else {
                    let skip = 0;
                    const limit = MAX_ORIGINAL_TXS_PER_REQUEST;
                    if (page) {
                        if (page < 1 || Number.isNaN(page)) {
                            reply.send(Crypto.sign({ success: false, error: `Invalid page number` }));
                            return;
                        }
                        skip = page - 1;
                        if (skip > 0)
                            skip = skip * limit;
                    }
                    originalTxs = await OriginalTxDB.queryOriginalTxsData(skip, limit, from, to);
                }
            }
            const res = Crypto.sign({
                originalTxs,
            });
            reply.send(res);
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('POST_originalTx');
        }
    });
    server.post('/receipt', async (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('POST_receipt');
        try {
            const requestData = _request.body;
            const result = (0, exports.validateRequestData)(requestData, {
                count: 'n?',
                start: 'n?',
                end: 'n?',
                startCycle: 'n?',
                endCycle: 'n?',
                type: 's?',
                page: 'n?',
                txId: 's?',
                txIdList: 'a?',
                sender: 's',
                sign: 'o',
            });
            if (!result.success) {
                reply.send(Crypto.sign({ success: false, error: result.error }));
                return;
            }
            const { count, start, end, startCycle, endCycle, type, page, txId, txIdList } = _request.body;
            let receipts = [];
            if (count) {
                if (count <= 0 || Number.isNaN(count)) {
                    reply.send(Crypto.sign({ success: false, error: `Invalid count` }));
                    return;
                }
                if (count > MAX_RECEIPTS_PER_REQUEST) {
                    reply.send(Crypto.sign({ success: false, error: `Max count is ${MAX_RECEIPTS_PER_REQUEST}` }));
                    return;
                }
                receipts = await ReceiptDB.queryLatestReceipts(count);
            }
            else if (txId) {
                if (txId.length !== TXID_LENGTH) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Invalid txId ${txId}`,
                    }));
                    return;
                }
                const receipt = await ReceiptDB.queryReceiptByReceiptId(txId);
                if (receipt)
                    receipts.push(receipt);
            }
            else if (txIdList) {
                if (txIdList.length > MAX_RECEIPTS_PER_REQUEST) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Exceed maximum limit of ${MAX_RECEIPTS_PER_REQUEST} receipts`,
                    }));
                    return;
                }
                for (const [txId, txTimestamp] of txIdList) {
                    if (typeof txId !== 'string' || txId.length !== TXID_LENGTH || typeof txTimestamp !== 'number') {
                        reply.send(Crypto.sign({
                            success: false,
                            error: `Invalid txId ${txId} in the List`,
                        }));
                        return;
                    }
                    const receipt = await ReceiptDB.queryReceiptByReceiptId(txId, txTimestamp);
                    if (receipt)
                        receipts.push(receipt);
                }
            }
            else if (start || start === 0) {
                const from = start;
                const to = end ? end : from;
                if (!(from >= 0 && to >= from) || Number.isNaN(from) || Number.isNaN(to)) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Invalid start and end counters`,
                    }));
                    return;
                }
                const count = to - from;
                if (count > MAX_RECEIPTS_PER_REQUEST) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Exceed maximum limit of ${MAX_RECEIPTS_PER_REQUEST} receipts`,
                    }));
                    return;
                }
                receipts = await ReceiptDB.queryReceipts(from, count + 1);
            }
            else if (startCycle || startCycle === 0) {
                const from = startCycle;
                const to = endCycle ? endCycle : from;
                if (!(from >= 0 && to >= from) || Number.isNaN(from) || Number.isNaN(to)) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Invalid startCycle and endCycle counters`,
                    }));
                    return;
                }
                const count = to - from;
                if (count > MAX_BETWEEN_CYCLES_PER_REQUEST) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Exceed maximum limit of ${MAX_BETWEEN_CYCLES_PER_REQUEST} cycles`,
                    }));
                    return;
                }
                if (type === 'tally') {
                    receipts = await ReceiptDB.queryReceiptCountByCycles(from, to);
                }
                else if (type === 'count') {
                    receipts = await ReceiptDB.queryReceiptCountBetweenCycles(from, to);
                }
                else {
                    let skip = 0;
                    const limit = MAX_RECEIPTS_PER_REQUEST;
                    if (page) {
                        if (page < 1 || Number.isNaN(page)) {
                            reply.send(Crypto.sign({ success: false, error: `Invalid page number` }));
                            return;
                        }
                        skip = page - 1;
                        if (skip > 0)
                            skip = skip * limit;
                    }
                    receipts = await ReceiptDB.queryReceiptsBetweenCycles(skip, limit, from, to);
                }
            }
            const res = Crypto.sign({
                receipts,
            });
            reply.send(res);
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('POST_receipt');
        }
    });
    server.post('/account', async (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('POST_account');
        try {
            const requestData = _request.body;
            const result = (0, exports.validateRequestData)(requestData, {
                count: 'n?',
                start: 'n?',
                end: 'n?',
                startCycle: 'n?',
                endCycle: 'n?',
                page: 'n?',
                accountId: 's?',
                sender: 's',
                sign: 'o',
            });
            if (!result.success) {
                reply.send(Crypto.sign({ success: false, error: result.error }));
                return;
            }
            let accounts = [];
            let totalAccounts = 0;
            let res;
            const { count, start, end, startCycle, endCycle, page, accountId } = _request.body;
            if (count) {
                if (count <= 0 || Number.isNaN(count)) {
                    reply.send(Crypto.sign({ success: false, error: `Invalid count` }));
                    return;
                }
                if (count > MAX_ACCOUNTS_PER_REQUEST) {
                    reply.send(Crypto.sign({ success: false, error: `Max count is ${MAX_ACCOUNTS_PER_REQUEST}` }));
                    return;
                }
                accounts = await AccountDB.queryLatestAccounts(count);
                res = { accounts };
            }
            else if (start || start === 0) {
                const from = start;
                const to = end ? end : from;
                if (!(from >= 0 && to >= from) || Number.isNaN(from) || Number.isNaN(to)) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Invalid start and end counters`,
                    }));
                    return;
                }
                const count = to - from;
                if (count > MAX_ACCOUNTS_PER_REQUEST) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Exceed maximum limit of ${MAX_ACCOUNTS_PER_REQUEST} accounts`,
                    }));
                    return;
                }
                accounts = await AccountDB.queryAccounts(from, count + 1);
                res = { accounts };
            }
            else if (startCycle || startCycle === 0) {
                const from = startCycle;
                const to = endCycle ? endCycle : from;
                if (!(from >= 0 && to >= from) || Number.isNaN(from) || Number.isNaN(to)) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Invalid startCycle and endCycle counters`,
                    }));
                    return;
                }
                const count = to - from;
                if (count > MAX_BETWEEN_CYCLES_PER_REQUEST) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Exceed maximum limit of ${MAX_BETWEEN_CYCLES_PER_REQUEST} cycles to query accounts Count`,
                    }));
                    return;
                }
                totalAccounts = await AccountDB.queryAccountCountBetweenCycles(from, to);
                if (page) {
                    if (page < 1 || Number.isNaN(page)) {
                        reply.send(Crypto.sign({ success: false, error: `Invalid page number` }));
                        return;
                    }
                    let skip = page - 1;
                    const limit = MAX_ACCOUNTS_PER_REQUEST;
                    if (skip > 0)
                        skip = skip * limit;
                    accounts = await AccountDB.queryAccountsBetweenCycles(skip, limit, from, to);
                    res = { accounts, totalAccounts };
                }
                else {
                    res = { totalAccounts };
                }
            }
            else if (accountId) {
                accounts = await AccountDB.queryAccountByAccountId(accountId);
                res = { accounts };
            }
            else {
                reply.send(Crypto.sign({
                    success: false,
                    error: 'not specified which account to show',
                }));
                return;
            }
            reply.send(Crypto.sign(res));
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('POST_account');
        }
    });
    server.post('/transaction', async (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('POST_transaction');
        try {
            const requestData = _request.body;
            const result = (0, exports.validateRequestData)(requestData, {
                count: 'n?',
                start: 'n?',
                end: 'n?',
                txId: 's?',
                appReceiptId: 's?',
                startCycle: 'n?',
                endCycle: 'n?',
                page: 'n?',
                sender: 's',
                sign: 'o',
            });
            if (!result.success) {
                reply.send(Crypto.sign({ success: false, error: result.error }));
                return;
            }
            const { count, start, end, txId, appReceiptId, startCycle, endCycle, page } = _request.body;
            let transactions = [];
            let totalTransactions = 0;
            let res;
            if (count) {
                if (count <= 0 || Number.isNaN(count)) {
                    reply.send(Crypto.sign({ success: false, error: `Invalid count` }));
                    return;
                }
                if (count > MAX_ACCOUNTS_PER_REQUEST) {
                    reply.send(Crypto.sign({ success: false, error: `Max count is ${MAX_ACCOUNTS_PER_REQUEST}` }));
                    return;
                }
                transactions = await TransactionDB.queryLatestTransactions(count);
                res = { transactions };
            }
            else if (start || start === 0) {
                const from = start;
                const to = end ? end : from;
                if (!(from >= 0 && to >= from) || Number.isNaN(from) || Number.isNaN(to)) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Invalid start and end counters`,
                    }));
                    return;
                }
                const count = to - from;
                if (count > MAX_ACCOUNTS_PER_REQUEST) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Exceed maximum limit of ${MAX_ACCOUNTS_PER_REQUEST} transactions`,
                    }));
                    return;
                }
                transactions = await TransactionDB.queryTransactions(from, count + 1);
                res = { transactions };
            }
            else if (startCycle || startCycle === 0) {
                const from = startCycle;
                const to = endCycle ? endCycle : from;
                if (!(from >= 0 && to >= from) || Number.isNaN(from) || Number.isNaN(to)) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Invalid startCycle and endCycle counters`,
                    }));
                    return;
                }
                const count = to - from;
                if (count > MAX_BETWEEN_CYCLES_PER_REQUEST) {
                    reply.send(Crypto.sign({
                        success: false,
                        error: `Exceed maximum limit of ${MAX_BETWEEN_CYCLES_PER_REQUEST} cycles to query transactions Count`,
                    }));
                    return;
                }
                totalTransactions = await TransactionDB.queryTransactionCountBetweenCycles(from, to);
                if (page) {
                    if (page < 1 || Number.isNaN(page)) {
                        reply.send(Crypto.sign({ success: false, error: `Invalid page number` }));
                        return;
                    }
                    let skip = page - 1;
                    const limit = MAX_ACCOUNTS_PER_REQUEST;
                    if (skip > 0)
                        skip = skip * limit;
                    transactions = await TransactionDB.queryTransactionsBetweenCycles(skip, limit, from, to);
                    res = { transactions, totalTransactions };
                }
                else {
                    res = { totalTransactions };
                }
            }
            else if (txId) {
                transactions = await TransactionDB.queryTransactionByTxId(txId);
                res = { transactions };
            }
            else if (appReceiptId) {
                transactions = await TransactionDB.queryTransactionByAccountId(appReceiptId);
                res = { transactions };
            }
            else {
                res = {
                    success: false,
                    error: 'not specified which transaction to show',
                };
                reply.send(Crypto.sign(res));
                return;
            }
            reply.send(Crypto.sign(res));
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('POST_transaction');
        }
    });
    server.post('/totalData', async (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('POST_totalData');
        try {
            const requestData = _request.body;
            const result = (0, exports.validateRequestData)(requestData, {
                sender: 's',
                sign: 'o',
            });
            if (!result.success) {
                reply.send(Crypto.sign({ success: false, error: result.error }));
                return;
            }
            const totalCycles = await CycleDB.queryCyleCount();
            const totalAccounts = await AccountDB.queryAccountCount();
            const totalTransactions = await TransactionDB.queryTransactionCount();
            const totalReceipts = await ReceiptDB.queryReceiptCount();
            const totalOriginalTxs = await OriginalTxDB.queryOriginalTxDataCount();
            // Get the last five minutes bucket status for each checkpoint manager
            const cycleLastFiveMinutesGiveUpBucketStatus = (0, Utils_1.getCheckpointManager)(CheckpointData_1.CheckpointType.Cycle)?.hasLastFailedBucketExceededDuration();
            const originalTxLastFiveMinutesGiveUpBucketStatus = (0, Utils_1.getCheckpointManager)(CheckpointData_1.CheckpointType.OriginalTx)?.hasLastFailedBucketExceededDuration();
            const receiptLastFiveMinutesGiveUpBucketStatus = (0, Utils_1.getCheckpointManager)(CheckpointData_1.CheckpointType.Receipt)?.hasLastFailedBucketExceededDuration();
            reply.send(Crypto.sign({
                totalCycles,
                totalAccounts,
                totalTransactions,
                totalReceipts,
                totalOriginalTxs,
                cycleLastFiveMinutesGiveUpBucketStatus,
                originalTxLastFiveMinutesGiveUpBucketStatus,
                receiptLastFiveMinutesGiveUpBucketStatus,
            }));
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('POST_totalData');
        }
    });
    server.post('/gossip-data', async (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('POST_gossip-data');
        try {
            const gossipPayload = _request.body;
            if (Config_1.config.VERBOSE)
                Logger.mainLogger.debug('Gossip Data received', lib_types_2.Utils.safeStringify(gossipPayload));
            const result = Collector.validateGossipData(gossipPayload);
            if (!result.success) {
                reply.send({ success: false, error: result.error });
                return;
            }
            const res = Crypto.sign({
                success: true,
            });
            reply.send(res);
            Collector.processGossipData(gossipPayload);
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('POST_gossip-data');
        }
    });
    server.post('/get_account_data_archiver', async (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('POST_get_account_data');
        try {
            const payload = _request.body;
            if (Config_1.config.VERBOSE)
                Logger.mainLogger.debug('Account Data received', lib_types_2.Utils.safeStringify(payload));
            const result = AccountDataProvider.validateAccountDataRequest(payload);
            // Logger.mainLogger.debug('Account Data validation result', result)
            if (!result.success) {
                reply.send({ success: false, error: result.error });
                return;
            }
            if (payload.maxRecords > Config_1.config.maxRecordsPerRequest || payload.maxRecords <= 0) {
                reply.send({
                    success: false,
                    error: `Invalid AccountBucket size. Size was ${payload.maxRecords}. Must be greater than 0 and less than ${Config_1.config.maxRecordsPerRequest}.`,
                });
                return;
            }
            const data = await AccountDataProvider.provideAccountDataRequest(payload);
            // Logger.mainLogger.debug('Account Data result', data)
            const res = Crypto.sign({
                success: true,
                data,
            });
            reply.send(res);
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('POST_get_account_data');
        }
    });
    server.post('/get_account_data_by_list_archiver', async (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('POST_get_account_data_by_list');
        try {
            const payload = _request.body;
            if (Config_1.config.VERBOSE)
                Logger.mainLogger.debug('Account Data By List received', lib_types_2.Utils.safeStringify(payload));
            const result = AccountDataProvider.validateAccountDataByListRequest(payload);
            // Logger.mainLogger.debug('Account Data By List validation result', result)
            if (!result.success) {
                reply.send({ success: false, error: result.error });
                return;
            }
            const accountData = await AccountDataProvider.provideAccountDataByListRequest(payload);
            // Logger.mainLogger.debug('Account Data By List result', accountData)
            const res = Crypto.sign({
                success: true,
                accountData,
            });
            reply.send(res);
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('POST_get_account_data_by_list');
        }
    });
    server.post('/get_globalaccountreport_archiver', async (_request, reply) => {
        profiler_1.profilerInstance.profileSectionStart('GET_globalaccountreport_archiver');
        try {
            const payload = _request.body;
            if (Config_1.config.VERBOSE)
                Logger.mainLogger.debug('Global Account Report received', lib_types_2.Utils.safeStringify(payload));
            const result = AccountDataProvider.validateGlobalAccountReportRequest(payload);
            // Logger.mainLogger.debug('Global Account Report validation result', result)
            if (!result.success) {
                reply.send({ success: false, error: result.error });
                return;
            }
            const report = await AccountDataProvider.provideGlobalAccountReportRequest();
            // Logger.mainLogger.debug('Global Account Report result', report)
            const res = Crypto.sign(report);
            reply.send(res);
        }
        finally {
            profiler_1.profilerInstance.profileSectionEnd('GET_globalaccountreport_archiver');
        }
    });
    /**
     * This endpoint is used to update the archiver config.
     * @script A test script to hit this endpoint is available at `scripts/update_config.ts`
     */
    server.patch('/set-config', {
        preHandler: async (_request, reply) => {
            (0, DebugMode_1.isDebugMiddleware)(_request, reply);
        },
    }, async (_request, reply) => {
        const RESTRICTED_PARAMS = [
            'ARCHIVER_IP',
            'ARCHIVER_PORT',
            'ARCHIVER_HASH_KEY',
            'ARCHIVER_SECRET_KEY',
            'ARCHIVER_PUBLIC_KEY',
        ];
        try {
            const { sign, ...newConfig } = _request.body;
            const validKeys = new Set(Object.keys(Config_1.config));
            const payloadKeys = Object.keys(newConfig);
            const invalidKeys = payloadKeys.filter((key) => !validKeys.has(key) || RESTRICTED_PARAMS.includes(key));
            if (invalidKeys.length > 0)
                throw new Error(`Invalid/Unauthorised config properties provided: ${invalidKeys.join(', ')}`);
            if (Config_1.config.VERBOSE)
                Logger.mainLogger.debug('Archiver config update executed: ', JSON.stringify(newConfig));
            const updatedConfig = (0, Config_1.updateConfig)(newConfig);
            reply.send({ success: true, ...updatedConfig, ARCHIVER_SECRET_KEY: '' });
        }
        catch (error) {
            reply.status(400).send({ success: false, reason: error.message });
        }
    });
    // Config Endpoint
    server.get('/config', {
        preHandler: async (_request, reply) => {
            (0, DebugMode_1.isDebugMiddleware)(_request, reply);
        },
    }, (_request, reply) => {
        reply.send({ ...Config_1.config, ARCHIVER_SECRET_KEY: '' }); // send the config without the secret key
    });
    // dataSenders Endpoint
    server.get('/dataSenders', {
        preHandler: async (_request, reply) => {
            (0, DebugMode_1.isDebugMiddleware)(_request, reply);
        },
    }, (_request, reply) => {
        const data = {
            dataSendersSize: Data.dataSenders.size,
            socketClientsSize: Data.socketClients.size,
        };
        if (_request.query && _request.query['dataSendersList'] === 'true')
            data['dataSendersList'] = Array.from(Data.dataSenders.values()).map((item) => item.nodeInfo.ip + ':' + item.nodeInfo.port);
        reply.send(data);
    });
    const enableLoseYourself = false; // set this to `true` during testing, but never commit as `true`
    server.get('/lose-yourself', 
    // this debug endpoint is in support of development and testing of
    // lost archiver detection
    {
        preHandler: async (_request, reply) => {
            (0, DebugMode_1.isDebugMiddleware)(_request, reply);
        },
    }, (_request, reply) => {
        if (enableLoseYourself) {
            Logger.mainLogger.debug('/lose-yourself: exit(1)');
            reply.send({ status: 'success', message: 'will exit' });
            // We don't call exitArchiver() here because that awaits Data.sendLeaveRequest(...),
            // but we're simulating a lost node.
            process.exit(1);
        }
        else {
            Logger.mainLogger.debug('/lose-yourself: not enabled. no action taken.');
            reply.send({ status: 'failure', message: 'not enabled' });
            // set enableLoseYourself to true--but never commit!
        }
    });
    // ping the archiver to see if it's alive
    server.get('/ping', 
    // this debug endpoint is in support of development and testing of
    // lost archiver detection
    {
        preHandler: async (_request, reply) => {
            (0, DebugMode_1.isDebugMiddleware)(_request, reply);
        },
    }, (request, reply) => {
        if (reachabilityAllowed) {
            reply.status(200).send('pong!');
        }
        else {
            request.raw.socket.destroy();
        }
    });
    server.get('/set-reachability', {
        schema: {
            querystring: {
                properties: {
                    value: {
                        type: 'boolean',
                    },
                },
            },
        },
        preHandler: async (request, reply) => {
            (0, DebugMode_1.isDebugMiddleware)(request, reply);
        },
    }, async (request, reply) => {
        const msg = `/set-reachability`;
        console.log(msg);
        Logger.mainLogger.info(msg);
        const value = request.query.value;
        if (typeof value !== 'boolean') {
            Logger.mainLogger.info('/set-reachability: value must be a boolean');
            reply.status(400).send('value must be a boolean');
        }
        else {
            const msg = `/set-reachability: ${value}`;
            console.log(msg);
            Logger.mainLogger.info(msg);
            reachabilityAllowed = value;
        }
    });
    server.get('/verified-receipt-counter', {
        preHandler: async (_request, reply) => {
            (0, DebugMode_1.isDebugMiddleware)(_request, reply);
        },
    }, (_request, reply) => {
        reply.send({
            receivedReceiptCount: primary_process_1.receivedReceiptCount,
            verifiedReceiptCount: primary_process_1.verifiedReceiptCount,
            successReceiptCount: primary_process_1.successReceiptCount,
            failureReceiptCount: primary_process_1.failureReceiptCount,
        });
    });
    server.get('/status', async (_request, reply) => {
        Logger.mainLogger.debug('Status request received');
        const status = {
            isActive: State.isActive,
        };
        Logger.mainLogger.debug('Status response', status);
        const res = Crypto.sign({
            status,
        });
        reply.send(res);
    });
    // Old snapshot ArchivedCycle endpoint;
    if (!Config_1.config.experimentalSnapshot) {
        server.get('/full-archive', async (_request, reply) => {
            const err = Utils.validateTypes(_request.query, { start: 's', end: 's' });
            if (err) {
                reply.send({ success: false, error: err });
                return;
            }
            const { start, end } = _request.query;
            const from = start ? parseInt(start) : 0;
            const to = end ? parseInt(end) : from;
            if (!(from >= 0 && to >= from) || Number.isNaN(from) || Number.isNaN(to)) {
                reply.send({ success: false, error: `Invalid start and end counters` });
                return;
            }
            const count = to - from;
            if (count > MAX_BETWEEN_CYCLES_PER_REQUEST) {
                reply.send({
                    success: false,
                    error: `Exceed maximum limit of ${MAX_BETWEEN_CYCLES_PER_REQUEST} cycles`,
                });
                return;
            }
            let archivedCycles = [];
            archivedCycles = await Storage.queryAllArchivedCyclesBetween(from, to);
            const res = Crypto.sign({
                archivedCycles,
            });
            reply.send(res);
        });
        server.get('/full-archive/:count', async (_request, reply) => {
            const err = Utils.validateTypes(_request.params, { count: 's' });
            if (err) {
                reply.send({ success: false, error: err });
                return;
            }
            const count = parseInt(_request.params.count);
            if (count <= 0 || Number.isNaN(count)) {
                reply.send({ success: false, error: `Invalid count` });
                return;
            }
            if (count > MAX_BETWEEN_CYCLES_PER_REQUEST) {
                reply.send({ success: false, error: `Max count is ${MAX_BETWEEN_CYCLES_PER_REQUEST}` });
                return;
            }
            const archivedCycles = await Storage.queryAllArchivedCycles(count);
            const res = Crypto.sign({
                archivedCycles,
            });
            reply.send(res);
        });
        server.post('/gossip-hashes', async (_request, reply) => {
            const gossipMessage = _request.body;
            Logger.mainLogger.debug('Gossip received', lib_types_2.Utils.safeStringify(gossipMessage));
            (0, Gossip_1.addHashesGossip)(gossipMessage.sender, gossipMessage.data);
            const res = Crypto.sign({
                success: true,
            });
            reply.send(res);
        });
    }
    server.get('/get-network-account', (_request, reply) => {
        const useHash = _request.query?.hash !== 'false';
        const response = useHash
            ? { networkAccountHash: (0, GlobalAccount_1.getGlobalNetworkAccount)(useHash) }
            : { networkAccount: (0, GlobalAccount_1.getGlobalNetworkAccount)(useHash) };
        // We might want to sign this response
        reply.send(Crypto.sign(response));
    });
    // Register ticket routes
    server.register(tickets_1.default, { prefix: '/tickets' });
    server.post('/shareCheckpointRadixDigests', async (req, reply) => {
        try {
            const result = (0, exports.validateRequestData)(req.body, {
                sender: 's',
                sign: 'o',
                senderAddress: 's',
                bucketID: 's',
                radixDigests: 's',
                checkpointType: 'n',
                startTime: 'n',
                endTime: 'n',
            });
            if (!result.success) {
                reply.send({ success: false, error: result.error });
                return;
            }
            const { bucketID, radixDigests, senderAddress, checkpointType } = req.body;
            // Identify the correct manager based on the checkpoint type
            const manager = (0, Utils_1.getCheckpointManager)(checkpointType);
            let bucket = manager.checkpointBuckets.get(bucketID);
            if (!bucket) {
                Logger.mainLogger.info(`missing bucket for bucketID=${bucketID}, checkpoint type ${checkpointType}`);
                nestedCounters_1.nestedCountersInstance.countEvent('checkpoint', `missing bucket for bucketID=${bucketID}, checkpoint type ${checkpointType}`);
                if (Config_1.config.VERBOSE) {
                    Logger.mainLogger.debug(`Bucket not found: No bucket with ID=${bucketID} for checkpoint type ${checkpointType}.`);
                }
                const startTime = req.body.startTime;
                const endTime = req.body.endTime;
                bucket = new CheckpointData_1.CheckpointBucket(startTime, endTime, bucketID, manager.validateData, manager.updateData, manager.checkpointType);
                manager.checkpointBuckets.set(bucketID, bucket);
                Logger.mainLogger.info(`Successfully created missing bucket with ID=${bucketID} for checkpoint type ${checkpointType}`);
            }
            // Process received digests
            Logger.mainLogger.info(`Processing received digests for bucket ID=${bucketID} from ${senderAddress}.`);
            manager.onHashDigestsReceived(senderAddress, bucketID, lib_types_2.Utils.safeJsonParse(radixDigests));
            reply.status(200).send({
                success: true,
            });
        }
        catch (err) {
            Logger.mainLogger.error(`Error processing shareCheckpointRadixDigests for checkpoint type ${req.body.checkpointType}: ${err.message}`);
            reply.status(500).send('Server error');
        }
    });
    server.post('/exchangeCheckpointRadixEntries', async (req, reply) => {
        try {
            const result = (0, exports.validateRequestData)(req.body, {
                sender: 's',
                sign: 'o',
                bucketID: 's',
                entries: 's',
                checkpointType: 'n',
            });
            if (!result.success) {
                reply.send({ success: false, error: result.error });
                return;
            }
            const { bucketID, entries, checkpointType } = req.body;
            const parsedEntries = lib_types_2.Utils.safeJsonParse(entries);
            const manager = (0, Utils_1.getCheckpointManager)(checkpointType);
            const bucket = manager.checkpointBuckets.get(bucketID);
            if (!bucket) {
                if (Config_1.config.VERBOSE) {
                    Logger.mainLogger.debug(`Bucket not found: No bucket with ID=${bucketID} for checkpoint type ${checkpointType}.`);
                }
                reply.status(404).send(`Bucket not found for ID=${bucketID}.`);
                return;
            }
            // Get our entries to share back
            const ourEntries = [];
            // For each incoming entry, get our corresponding entry
            for (const incomingEntry of parsedEntries) {
                const radix = incomingEntry.digest.radix;
                const ourEntry = bucket.radixEntries.get(radix);
                if (ourEntry) {
                    ourEntry.updateDigest();
                    ourEntries.push(ourEntry);
                }
            }
            // Process their entries
            bucket.onExchangeRadixEntries(bucketID, parsedEntries);
            // Send our entries back
            const res = Crypto.sign({
                bucketID,
                entries: ourEntries,
                success: true,
            });
            reply.send(res);
        }
        catch (err) {
            Logger.mainLogger.error(`Error processing exchangeCheckpointRadixEntries for checkpoint type ${req.body.checkpointType}: ${err.message}`);
            reply.status(500).send(`Server error in exchangeCheckpointRadixEntries for type ${req.body.checkpointType}`);
        }
    });
    server.post('/bucket-verification', async (request, reply) => {
        try {
            const requestData = request.body;
            const result = (0, exports.validateRequestData)(requestData, {
                bucketID: 's',
                endBucketID: '?s',
                sender: 's',
                sign: 'o',
            });
            if (!result.success) {
                reply.code(400).send({
                    success: false,
                    error: result.error,
                });
                return;
            }
            const bucketID = parseInt(requestData.bucketID, 10);
            let endBucketID = undefined;
            if (requestData.endBucketID) {
                endBucketID = parseInt(requestData.endBucketID, 10);
                if (isNaN(endBucketID) || endBucketID < 0) {
                    reply.code(400).send({
                        success: false,
                        error: 'Invalid endBucketID. Must be a non-negative integer.',
                    });
                    return;
                }
            }
            if (isNaN(bucketID) || bucketID < 0) {
                reply.code(400).send({
                    success: false,
                    error: 'Invalid bucketID. Must be a non-negative integer.',
                });
                return;
            }
            const isVerified = await (0, checkpointStatus_1.isBucketVerified)(bucketID, endBucketID);
            const response = Crypto.sign({
                success: true,
                isVerified,
            });
            reply.send(response);
        }
        catch (error) {
            Logger.mainLogger.error(`Error checking bucket verification status:`, error);
            reply.code(500).send({
                success: false,
                error: 'Internal server error while checking bucket verification status',
            });
        }
    });
}
exports.registerRoutes = registerRoutes;
const validateRequestData = (data, expectedDataType, skipArchiverCheck = false) => {
    try {
        let err = Utils.validateTypes(data, expectedDataType);
        if (err) {
            Logger.mainLogger.error('Invalid request data ', err);
            return { success: false, error: 'Invalid request data ' + err };
        }
        err = Utils.validateTypes(data.sign, { owner: 's', sig: 's' });
        if (err) {
            Logger.mainLogger.error('Invalid request data signature ', err);
            return { success: false, error: 'Invalid request data signature ' + err };
        }
        if (data.sign.owner !== data.sender) {
            Logger.mainLogger.error('Data sender publicKey and sign owner key does not match');
            return { success: false, error: 'Data sender publicKey and sign owner key does not match' };
        }
        if (!skipArchiverCheck && Config_1.config.limitToArchiversOnly) {
            // Check if the sender is in the allowed archivers list
            const isAllowedArchiver = allowedArchiversManager_1.allowedArchiversManager.isArchiverAllowed(data.sender);
            // Check if the sender is in the active archiver list or is the devPublicKey
            const isActiveArchiver = State.activeArchivers.some((archiver) => archiver.publicKey === data.sender);
            const approvedSender = isAllowedArchiver || (isAllowedArchiver && isActiveArchiver) || Config_1.config.DevPublicKey === data.sender;
            if (!approvedSender) {
                return {
                    success: false,
                    error: isAllowedArchiver
                        ? 'Archiver is not in the allowed archiver list'
                        : 'Data request sender is not an authorized archiver',
                };
            }
        }
        if (!Crypto.verify(data)) {
            Logger.mainLogger.error('Invalid signature', data);
            return { success: false, error: 'Invalid signature' };
        }
        return { success: true };
    }
    catch (e) {
        Logger.mainLogger.error('Error validating request data', e);
        return { success: false, error: 'Error validating request data' };
    }
};
exports.validateRequestData = validateRequestData;
var RequestDataType;
(function (RequestDataType) {
    RequestDataType["CYCLE"] = "cycleinfo";
    RequestDataType["RECEIPT"] = "receipt";
    RequestDataType["ORIGINALTX"] = "originalTx";
    RequestDataType["ACCOUNT"] = "account";
    RequestDataType["TRANSACTION"] = "transaction";
    RequestDataType["TOTALDATA"] = "totalData";
    RequestDataType["CHECKPOINT"] = "checkpoint";
})(RequestDataType = exports.RequestDataType || (exports.RequestDataType = {}));
const queryFromArchivers = async (type, queryParameters, timeoutInSecond) => {
    let bucketID;
    const params = queryParameters;
    let checkpointStatusType;
    let bucketRange;
    // Extract bucket ID based on the request type and parameters
    switch (type) {
        case RequestDataType.CYCLE:
            // For cycle info, use start or end as bucket ID
            if (params.start !== undefined && params.end !== undefined) {
                // If both start and end are provided, we'll check the entire range
                bucketRange = {
                    start: params.start,
                    end: params.end,
                };
                bucketID = params.start; // Default to start for single bucket check fallback
            }
            else if (params.start !== undefined) {
                bucketID = params.start;
            }
            else if (params.end !== undefined) {
                bucketID = params.end;
            }
            checkpointStatusType = checkpointStatus_1.CheckpointStatusType.CYCLE;
            break;
        case RequestDataType.RECEIPT:
            checkpointStatusType = checkpointStatus_1.CheckpointStatusType.RECEIPT;
            break;
        case RequestDataType.ORIGINALTX:
            checkpointStatusType = checkpointStatus_1.CheckpointStatusType.ORIGINAL_TX;
            break;
        case RequestDataType.ACCOUNT:
        case RequestDataType.TRANSACTION:
            // For other data types, check if startCycle or endCycle is specified
            if (params.startCycle !== undefined && params.endCycle !== undefined) {
                // If both startCycle and endCycle are provided, we'll check the entire range
                bucketRange = {
                    start: params.startCycle,
                    end: params.endCycle,
                };
                bucketID = params.startCycle; // Default to startCycle for single bucket check fallback
            }
            else if (params.startCycle !== undefined) {
                bucketID = params.startCycle;
            }
            else if (params.endCycle !== undefined) {
                bucketID = params.endCycle;
            }
            break;
    }
    const data = {
        ...queryParameters,
        sender: Config_1.config.ARCHIVER_PUBLIC_KEY,
    };
    const signedDataToSend = Crypto.sign(data);
    let url;
    switch (type) {
        case RequestDataType.CYCLE:
            url = `/cycleinfo`;
            break;
        case RequestDataType.RECEIPT:
            url = `/receipt`;
            break;
        case RequestDataType.ORIGINALTX:
            url = `/originalTx`;
            break;
        case RequestDataType.ACCOUNT:
            url = `/account`;
            break;
        case RequestDataType.TRANSACTION:
            url = `/transaction`;
            break;
        case RequestDataType.TOTALDATA:
            url = `/totalData`;
            break;
    }
    const maxNumberofArchiversToRetry = 3;
    const randomArchivers = Utils.getRandomItemFromArr(State.otherArchivers, 0, maxNumberofArchiversToRetry);
    // Try to find an archiver with verified data if checkpoint is enabled and we have bucket information
    if (Config_1.config.checkpoint.bucketConfig.allowCheckpointUpdates &&
        randomArchivers.length > 0 &&
        (bucketRange !== undefined || bucketID !== undefined)) {
        // Try to find an archiver with verified data for the bucket range or single bucket
        const verifiedArchiver = await findVerifiedArchiver(randomArchivers, bucketRange, bucketID, timeoutInSecond);
        if (verifiedArchiver) {
            try {
                const response = await P2P.postJson(`http://${verifiedArchiver.ip}:${verifiedArchiver.port}${url}`, signedDataToSend, timeoutInSecond);
                // Validate response using Crypto.verify
                if (response) {
                    const signatureVerificationResult = Crypto.verify(response);
                    if (signatureVerificationResult) {
                        return response;
                    }
                    else {
                        if (Config_1.config.VERBOSE) {
                            Logger.mainLogger.error(`Invalid response format from verified archiver ${verifiedArchiver.ip}:${verifiedArchiver.port} for ${type}`);
                        }
                    }
                }
            }
            catch (error) {
                Logger.mainLogger.error(`Error querying data from verified archiver ${verifiedArchiver.ip}:${verifiedArchiver.port}:`, error);
            }
        }
    }
    // If no archiver had verified data or checkpoint is disabled, try any archiver
    for (let retry = 0; retry < maxNumberofArchiversToRetry; retry++) {
        const randomArchiver = randomArchivers[retry] || randomArchivers[0];
        if (!randomArchiver)
            continue;
        try {
            const response = await P2P.postJson(`http://${randomArchiver.ip}:${randomArchiver.port}${url}`, signedDataToSend, timeoutInSecond);
            // Validate response using Crypto.verify
            if (response) {
                const signatureVerificationResult = Crypto.verify(response);
                if (signatureVerificationResult) {
                    return response;
                }
                else {
                    Logger.mainLogger.error(`Invalid response format from archiver ${randomArchiver.ip}:${randomArchiver.port} for ${type}`);
                }
            }
        }
        catch (e) {
            Logger.mainLogger.error(`Error while querying ${randomArchiver.ip}:${randomArchiver.port}${url} for data ${JSON.stringify(queryParameters)}`, e);
        }
    }
    return null;
};
exports.queryFromArchivers = queryFromArchivers;
// Helper function to find an archiver with verified data
async function findVerifiedArchiver(archivers, bucketRange, singleBucketID, timeoutInSecond) {
    for (const archiver of archivers) {
        if (!archiver)
            continue;
        try {
            if (bucketRange) {
                // Check verification for the bucket range
                const verificationData = {
                    bucketID: bucketRange.start.toString(),
                    endBucketID: bucketRange.end.toString(),
                    sender: Config_1.config.ARCHIVER_PUBLIC_KEY,
                };
                const signedBucketVerificationDataToSend = Crypto.sign(verificationData);
                const verificationResponse = await P2P.postJson(`http://${archiver.ip}:${archiver.port}/bucket-verification`, signedBucketVerificationDataToSend, timeoutInSecond);
                if (verificationResponse?.success && verificationResponse?.isVerified) {
                    const signatureVerificationResult = Crypto.verify(verificationResponse);
                    if (signatureVerificationResult) {
                        return archiver;
                    }
                }
            }
            else if (singleBucketID !== undefined) {
                // Check verification for a single bucket
                const verificationData = {
                    bucketID: singleBucketID.toString(),
                    sender: Config_1.config.ARCHIVER_PUBLIC_KEY,
                };
                const signedBucketVerificationDataToSend = Crypto.sign(verificationData);
                const verificationResponse = await P2P.postJson(`http://${archiver.ip}:${archiver.port}/bucket-verification`, signedBucketVerificationDataToSend, timeoutInSecond);
                if (verificationResponse?.success && verificationResponse?.isVerified) {
                    const signatureVerificationResult = Crypto.verify(verificationResponse);
                    if (signatureVerificationResult) {
                        return archiver;
                    }
                }
            }
        }
        catch (error) {
            Logger.mainLogger.error(`Error checking bucket verification for archiver ${archiver.ip}:${archiver.port}:`, error);
        }
    }
    return null;
}
//# sourceMappingURL=API.js.map