"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.setLastCycleToSync = exports.setSyncing = exports.getRandomArchiver = exports.setActive = exports.getCurvePk = exports.getCurveSk = exports.getSecretKey = exports.getNodeInfo = exports.compareCycleRecordWithOtherArchivers = exports.updateOtherArchivers = exports.resetActiveArchivers = exports.removeActiveArchiver = exports.addArchiver = exports.addSigListeners = exports.exitArchiver = exports.initFromConfig = exports.archiversReputation = exports.lastCycleToSync = exports.isSyncing = exports.isActive = exports.isFirst = exports.otherArchivers = exports.activeArchiversByPublicKeySorted = exports.activeArchivers = exports.joinedArchivers = void 0;
const Config_1 = require("./Config");
const Crypto = require("./Crypto");
const P2P = require("./P2P");
const NodeList = require("./NodeList");
const Data = require("./Data/Data");
const Utils = require("./Utils");
const Logger = require("./Logger");
const lib_archiver_discovery_1 = require("@shardeum-foundation/lib-archiver-discovery");
const GossipData_1 = require("./Data/GossipData");
const dbstore_1 = require("./dbstore");
const allowedArchiversManager_1 = require("./shardeum/allowedArchiversManager");
const customHttpFunctions_1 = require("./utils/customHttpFunctions");
const nodeState = {
    ip: '',
    port: -1,
    publicKey: '',
    secretKey: '',
    curvePk: '',
    curveSk: '',
};
exports.joinedArchivers = []; // Add joined archivers to this list first and move to activeArchivers when they are active
exports.activeArchivers = [];
exports.activeArchiversByPublicKeySorted = [];
// archivers list without the current archiver
exports.otherArchivers = [];
exports.isFirst = false;
exports.isActive = false;
exports.isSyncing = false;
exports.lastCycleToSync = 0;
exports.archiversReputation = new Map();
async function initFromConfig(config, shutDownMode = false, useArchiverDiscovery = true) {
    // Get own nodeInfo from config
    nodeState.ip = config.ARCHIVER_IP;
    nodeState.port = config.ARCHIVER_PORT;
    nodeState.publicKey = config.ARCHIVER_PUBLIC_KEY;
    nodeState.secretKey = config.ARCHIVER_SECRET_KEY;
    nodeState.curvePk = Crypto.core.convertPkToCurve(nodeState.publicKey);
    nodeState.curveSk = Crypto.core.convertSkToCurve(nodeState.secretKey);
    if (useArchiverDiscovery === false)
        return;
    let existingArchivers = [];
    // Parse existing archivers list
    try {
        console.log('ARCHIVER_INFO', process.env.ARCHIVER_INFO);
        console.log('Getting existing archivers list from archiver-discovery.');
        existingArchivers = (0, lib_archiver_discovery_1.getFinalArchiverList)().map(({ ip, port, publicKey }) => ({
            ip,
            port,
            publicKey,
            curvePk: Crypto.getOrCreateSharedKey(publicKey),
        }));
        console.log(`Got existing archivers list using archiver-discovery. [count: ${existingArchivers.length}]`);
    }
    catch (e) {
        console.warn('No existing archivers were found:', e);
    }
    if (existingArchivers.length === 0) {
        console.log('No existing archivers were found. This is the first archiver.');
        exports.isFirst = true;
        return;
    }
    if (shutDownMode)
        return;
    let retryCount = 1;
    const waitTime = 1000 * 60;
    while (retryCount < 10 && exports.activeArchivers.length === 0) {
        Logger.mainLogger.debug(`Getting consensor list from other achivers. [round: ${retryCount}]`);
        /* eslint-disable security/detect-object-injection */
        for (let i = 0; i < existingArchivers.length; i++) {
            if (existingArchivers[i].publicKey === nodeState.publicKey) {
                continue;
            }
            const response = (await P2P.getJson(`http://${existingArchivers[i].ip}:${existingArchivers[i].port}/nodelist`));
            Logger.mainLogger.debug('response', `http://${existingArchivers[i].ip}:${existingArchivers[i].port}/nodelist`, response);
            if (!response) {
                Logger.mainLogger.warn(`No response when fetching from archiver ${existingArchivers[i].ip}:${existingArchivers[i].port}`);
                continue;
            }
            if (!Crypto.verify(response)) {
                /* prettier-ignore */ console.log(`Invalid signature when fetching from archiver ${existingArchivers[i].ip}:${existingArchivers[i].port}`);
                continue;
            }
            if (response && response.nodeList && response.nodeList.length > 0) {
                addArchiver(existingArchivers[i]);
            }
        }
        /* eslint-enable security/detect-object-injection */
        if (exports.activeArchivers.length === 0) {
            Logger.mainLogger.error(`Unable to find active archivers. Waiting for ${waitTime} before trying again.`);
            // wait for 1 min before retrying
            await Utils.sleep(waitTime);
            retryCount += 1;
        }
    }
    if (exports.activeArchivers.length === 0) {
        Logger.mainLogger.error(`We have tried ${retryCount} times to get nodeList from other archivers. But got no response or empty list. About to exit now.`);
        process.exit(0);
    }
}
exports.initFromConfig = initFromConfig;
async function exitArchiver() {
    try {
        const randomConsensors = NodeList.getRandomActiveNodes(5);
        if (randomConsensors && randomConsensors.length > 0) {
            // Send a leave request to some random consensus nodes from the nodelist
            await Data.sendLeaveRequest(randomConsensors);
        }
        Logger.mainLogger.debug('Archiver will exit in 3 seconds.');
        setTimeout(() => {
            allowedArchiversManager_1.allowedArchiversManager.stopWatching();
            process.exit();
        }, 3000);
    }
    catch (e) {
        Logger.mainLogger.error(e);
    }
}
exports.exitArchiver = exitArchiver;
function addSigListeners(sigint = true, sigterm = true) {
    if (sigint) {
        process.on('SIGINT', async () => {
            Logger.mainLogger.debug('Exiting on SIGINT', process.pid);
            await (0, dbstore_1.closeDatabase)();
            allowedArchiversManager_1.allowedArchiversManager.stopWatching();
            if (exports.isActive)
                exitArchiver();
            else
                process.exit(0);
        });
    }
    if (sigterm) {
        process.on('SIGTERM', async () => {
            Logger.mainLogger.debug('Exiting on SIGTERM', process.pid);
            await (0, dbstore_1.closeDatabase)();
            allowedArchiversManager_1.allowedArchiversManager.stopWatching();
            if (exports.isActive)
                exitArchiver();
            else
                process.exit(0);
        });
    }
    Logger.mainLogger.debug('Registered exit signal listeners.');
}
exports.addSigListeners = addSigListeners;
function addArchiver(archiver) {
    const foundArchiver = exports.activeArchivers.find((a) => a.publicKey === archiver.publicKey);
    if (!foundArchiver) {
        Logger.mainLogger.debug('Adding archiver', archiver);
        exports.activeArchivers.push(archiver);
        Utils.insertSorted(exports.activeArchiversByPublicKeySorted, archiver, NodeList.byAscendingPublicKey);
        Logger.mainLogger.debug('activeArchiversByPublicKeySorted', exports.activeArchiversByPublicKeySorted.map((archiver) => archiver.publicKey));
        if (archiver.publicKey !== Config_1.config.ARCHIVER_PUBLIC_KEY) {
            exports.otherArchivers.push(archiver);
        }
        Logger.mainLogger.debug('New archiver added to active list', archiver);
    }
    Logger.mainLogger.debug('archivers list', exports.activeArchivers);
}
exports.addArchiver = addArchiver;
function removeActiveArchiver(publicKey) {
    exports.activeArchivers = exports.activeArchivers.filter((a) => a.publicKey !== publicKey);
    exports.activeArchiversByPublicKeySorted = exports.activeArchiversByPublicKeySorted.filter((a) => a.publicKey !== publicKey);
    exports.otherArchivers = exports.otherArchivers.filter((a) => a.publicKey !== publicKey);
    exports.archiversReputation.delete(publicKey);
}
exports.removeActiveArchiver = removeActiveArchiver;
function resetActiveArchivers(archivers) {
    Logger.mainLogger.debug('Resetting active archivers.', archivers);
    exports.activeArchivers = archivers;
    exports.activeArchiversByPublicKeySorted = [...archivers.sort(NodeList.byAscendingPublicKey)];
    exports.otherArchivers = exports.activeArchivers.filter((a) => a.publicKey !== Config_1.config.ARCHIVER_PUBLIC_KEY);
    exports.archiversReputation.clear();
    for (const archiver of exports.activeArchivers) {
        exports.archiversReputation.set(archiver.publicKey, 'up');
    }
    (0, GossipData_1.getAdjacentLeftAndRightArchivers)();
}
exports.resetActiveArchivers = resetActiveArchivers;
/**
 * Synchronizes the otherArchivers list to only include archivers that are present in activeArchivers,
 * excluding the current archiver (itself).
 */
function updateOtherArchivers() {
    // Filter out any archiver that is not in activeArchivers and is not the current archiver
    exports.otherArchivers = exports.activeArchivers.filter((a) => a.publicKey !== Config_1.config.ARCHIVER_PUBLIC_KEY);
}
exports.updateOtherArchivers = updateOtherArchivers;
async function compareCycleRecordWithOtherArchivers(archivers, ourCycleRecord) {
    // Remove after testing
    Logger.mainLogger.debug('[restore-409] activeArchivers: ', JSON.stringify(exports.activeArchivers));
    const promises = archivers.map((archiver) => (0, customHttpFunctions_1.customFetch)(`http://${archiver.ip}:${archiver.port}/cycleinfo/1`, {
        method: 'get',
        headers: { 'Content-Type': 'application/json' },
        timeout: 2000,
    }).then((res) => res.json()));
    let foundMatch = true;
    await Promise.allSettled(promises)
        .then((responses) => {
        let i = 0;
        for (const response of responses) {
            const archiver = exports.activeArchivers[i]; // eslint-disable-line security/detect-object-injection
            if (response.status === 'fulfilled') {
                const res = response.value;
                if (res && res.cycleInfo && res.cycleInfo.length > 0) {
                    const cycleInfo = res.cycleInfo[0];
                    if (cycleInfo.counter > ourCycleRecord.counter || cycleInfo.mode !== ourCycleRecord.mode) {
                        Logger.mainLogger.debug(`Our cycle record does not match with archiver ${archiver?.ip}:${archiver?.port}`);
                        // If our cycle record does not match with any of the archivers, maybe we are behind the network
                        foundMatch = false;
                        break;
                    }
                }
            }
            i++;
        }
    })
        .catch((error) => {
        // Handle any errors that occurred
        console.error(error);
    });
    return foundMatch;
}
exports.compareCycleRecordWithOtherArchivers = compareCycleRecordWithOtherArchivers;
function getNodeInfo() {
    const sanitizedNodeInfo = { ...nodeState };
    delete sanitizedNodeInfo.secretKey;
    delete sanitizedNodeInfo.curveSk;
    return sanitizedNodeInfo;
}
exports.getNodeInfo = getNodeInfo;
function getSecretKey() {
    return nodeState.secretKey;
}
exports.getSecretKey = getSecretKey;
function getCurveSk() {
    return nodeState.curveSk;
}
exports.getCurveSk = getCurveSk;
function getCurvePk() {
    return nodeState.curvePk;
}
exports.getCurvePk = getCurvePk;
function setActive() {
    exports.isActive = true;
}
exports.setActive = setActive;
function getRandomArchiver() {
    const randomArchiver = Utils.getRandomItemFromArr(exports.otherArchivers)[0];
    return randomArchiver;
}
exports.getRandomArchiver = getRandomArchiver;
function setSyncing(syncing) {
    exports.isSyncing = syncing;
}
exports.setSyncing = setSyncing;
function setLastCycleToSync(cycle) {
    exports.lastCycleToSync = cycle;
}
exports.setLastCycleToSync = setLastCycleToSync;
//# sourceMappingURL=State.js.map