export declare enum CheckpointStatusType {
    CYCLE = "cycle",
    RECEIPT = "receipt",
    ORIGINAL_TX = "original_tx"
}
export interface CheckpointStatus {
    cycle: number;
    unifiedStatus: boolean;
    cycleStatus: boolean;
    receiptStatus: boolean;
    originalTxStatus: boolean;
    created_at: number;
}
export declare const fieldMapping: Record<CheckpointStatusType, keyof CheckpointStatus>;
/**
 * Inserts or updates a checkpoint status record
 * @param status The checkpoint status to insert or update
 */
export declare function upsertCheckpointStatus(status: CheckpointStatus): Promise<void>;
/**
 * Updates a specific status field for a checkpoint
 * @param cycle The cycle number
 * @param statusField The status field to update ('cycleStatus', 'receiptStatus', or 'originalTxStatus')
 * @param value The boolean value to set
 */
export declare function updateCheckpointStatusField(cycle: number, statusField: CheckpointStatusType, value: boolean): Promise<void>;
/**
 * Bulk updates a specific status field (e.g., ORIGINAL_TX) for all cycles in [startCycle, endCycle].
 * This approach:
 *   - Grabs existing rows in one SELECT (so we know other status fields).
 *   - Updates/creates rows in memory.
 *   - Uses one multi-row upsert to write them back.
 */
export declare function bulkUpdateCheckpointStatusField(statusField: CheckpointStatusType, value: boolean, startCycle?: number, endCycle?: number, cycles?: number[]): Promise<void>;
/**
 * Gets a checkpoint status by cycle
 * @param cycle The cycle number
 */
export declare function getCheckpointStatus(cycle: number): Promise<CheckpointStatus | null>;
/**
 * Gets all checkpoint statuses with a specific unified status
 * @param unified Whether to get checkpoints with unified status true or false
 */
export declare function getCheckpointStatusesByUnifiedStatus(unified: boolean): Promise<CheckpointStatus[]>;
/**
 * Gets the oldest pending or failed checkpoint status
 * @returns The oldest pending or failed checkpoint status or null if none found
 */
export declare function getOldestPendingOrFailedCheckpointStatus(): Promise<CheckpointStatus | null>;
/**
 * Gets the range of cycles that need syncing
 * @returns Object with min and max cycle numbers that need syncing
 */
export declare function getCheckpointSyncRange(): Promise<{
    minCycle: number;
    maxCycle: number;
} | null>;
export declare function isBucketVerified(bucketID: number, endBucketID?: number): Promise<boolean>;
