"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const transactionVerification_1 = require("./services/transactionVerification");
const _startingMessage = `@shardeum-foundation/archiver starting at
  locale:  ${new Date().toLocaleString()}
  ISO/UTC: ${new Date().toISOString()}`;
console.log(_startingMessage);
console.error(_startingMessage);
const path_1 = require("path");
const fastify_1 = require("fastify");
const cors_1 = require("@fastify/cors");
const rate_limit_1 = require("@fastify/rate-limit");
const clusterModule = require("cluster");
const Config_1 = require("./Config");
const Crypto = require("./Crypto");
const State = require("./State");
const NodeList = require("./NodeList");
const Storage = require("./archivedCycle/Storage");
const Data = require("./Data/Data");
const Cycles = require("./Data/Cycles");
const DataLogWriter_1 = require("./Data/DataLogWriter");
const Utils = require("./Utils");
const StateMetaData_1 = require("./archivedCycle/StateMetaData");
const Logger = require("./Logger");
const fs_1 = require("fs");
const path_2 = require("path");
const memoryReporting_1 = require("./profiler/memoryReporting");
const nestedCounters_1 = require("./profiler/nestedCounters");
const profiler_1 = require("./profiler/profiler");
const statistics_1 = require("./statistics");
const dbstore = require("./dbstore");
const CycleDB = require("./dbstore/cycles");
const ReceiptDB = require("./dbstore/receipts");
const saveConsoleOutput_1 = require("./saveConsoleOutput");
const lib_archiver_discovery_1 = require("@shardeum-foundation/lib-archiver-discovery");
const Collector = require("./Data/Collector");
const GlobalAccount_1 = require("./GlobalAccount");
const Cycles_1 = require("./Data/Cycles");
const API_1 = require("./API");
const lib_types_1 = require("@shardeum-foundation/lib-types");
const healthCheck_1 = require("./routes/healthCheck");
const tickets_1 = require("./routes/tickets");
const Helpers_1 = require("./types/ajv/Helpers");
const SchemaHelpers_1 = require("./utils/serialization/SchemaHelpers");
const allowedArchiversManager_1 = require("./shardeum/allowedArchiversManager");
const CycleData_1 = require("./checkpoint/CycleData");
const ReceiptData_1 = require("./checkpoint/ReceiptData");
const OriginalTxsData_1 = require("./checkpoint/OriginalTxsData");
const Utils_1 = require("./Utils");
const CheckpointV2_1 = require("./checkpoint/CheckpointV2");
const checkpointStatus_1 = require("./dbstore/checkpointStatus");
const archiverLogging_1 = require("./profiler/archiverLogging");
const configFile = (0, path_2.resolve)(__dirname, '../archiver-config.json');
const allowedArchiversConfigPath = (0, path_1.join)(__dirname, '../allowed-archivers.json');
let logDir;
const cluster = clusterModule;
async function start() {
    (0, Config_1.overrideDefaultConfig)(configFile);
    (0, Helpers_1.initAjvSchemas)();
    (0, SchemaHelpers_1.initializeSerialization)();
    // Set crypto hash keys from config
    const hashKey = Config_1.config.ARCHIVER_HASH_KEY;
    Crypto.setCryptoHashKey(hashKey);
    let logsConfig;
    try {
        logsConfig = lib_types_1.Utils.safeJsonParse((0, fs_1.readFileSync)((0, path_2.resolve)(__dirname, '../archiver-log.json'), 'utf8'));
    }
    catch (err) {
        console.log('Failed to parse archiver log file:', err);
    }
    logDir = `${Config_1.config.ARCHIVER_LOGS}/${Config_1.config.ARCHIVER_IP}_${Config_1.config.ARCHIVER_PORT}`;
    const baseDir = '.';
    logsConfig.dir = logDir;
    Logger.initLogger(baseDir, logsConfig);
    if (logsConfig.saveConsoleOutput) {
        (0, saveConsoleOutput_1.startSaving)((0, path_1.join)(baseDir, logsConfig.dir));
    }
    // Initialize allowed archivers manager
    allowedArchiversManager_1.allowedArchiversManager.initialize(allowedArchiversConfigPath);
    // Global error handling
    process.on('uncaughtException', (error) => {
        Logger.mainLogger.error('Uncaught Exception - Global:', error);
    });
    process.on('unhandledRejection', (reason, promise) => {
        Logger.mainLogger.error('Unhandled Rejection - Global:', promise, 'reason:', reason);
    });
    // Initialize storage and checkpoints
    if (Config_1.config.experimentalSnapshot) {
        await dbstore.initializeDB(Config_1.config);
    }
    else {
        await Storage.initStorage(Config_1.config);
    }
    State.addSigListeners();
    if (!cluster.isPrimary) {
        // Initialize state from config
        await State.initFromConfig(Config_1.config, false, false);
        // await initWorkerProcess()
        return;
    }
    try {
        await (0, lib_archiver_discovery_1.setupArchiverDiscovery)({
            hashKey,
            customConfigPath: configFile.toString(),
            archiverTimeoutInMilliSeconds: 2000, // 2 seconds
        });
    }
    catch (e) {
        console.log('Error setting up archiver discovery: ', e);
    }
    const lastStoredCycle = await CycleDB.queryLatestCycleRecords(1);
    if (lastStoredCycle && lastStoredCycle.length > 0) {
        const lastStoredCycleMode = lastStoredCycle[0].mode;
        if (lastStoredCycleMode === 'shutdown') {
            (0, Cycles_1.setShutdownCycleRecord)(lastStoredCycle[0]);
            Logger.mainLogger.debug('Found shutdown cycleRecord', Cycles_1.cycleRecordWithShutDownMode);
            // Initialize state from config
            await State.initFromConfig(Config_1.config, true);
            const result = await State.compareCycleRecordWithOtherArchivers(Cycles_1.cycleRecordWithShutDownMode.archiversAtShutdown, Cycles_1.cycleRecordWithShutDownMode);
            if (result) {
                State.resetActiveArchivers(Cycles_1.cycleRecordWithShutDownMode.archiversAtShutdown);
                // Load global account from db
                await (0, GlobalAccount_1.loadGlobalAccounts)();
                await startServer();
                return;
            }
        }
    }
    // Initialize state from config
    await State.initFromConfig(Config_1.config);
    if (State.isFirst) {
        Logger.mainLogger.debug('We are first archiver. Starting archive-server');
        const lastStoredCycle = await CycleDB.queryLatestCycleRecords(1);
        if (lastStoredCycle && lastStoredCycle.length > 0) {
            archiverLogging_1.ArchiverLogging.logArchiverRegistration({
                archiverId: Config_1.config.ARCHIVER_IP,
                timestamp: Date.now(),
                validators: {
                    discovered: 0,
                    connected: 0,
                },
                state: 'REGISTERING',
            });
            // Load global account from db
            await (0, GlobalAccount_1.loadGlobalAccounts)();
            // Seems you got restarted, and there are no other archivers to check; build nodelists and send join request to the nodes first
            await Data.buildNodeListFromStoredCycle(lastStoredCycle[0]);
            let isJoined = false;
            let firstTime = true;
            const cycleDuration = Cycles.currentCycleDuration;
            do {
                try {
                    // Get active nodes from Archiver
                    const nodeList = NodeList.getActiveList();
                    // try to join the network
                    isJoined = await Data.joinNetwork(nodeList, firstTime);
                    if (isJoined) {
                        archiverLogging_1.ArchiverLogging.logArchiverRegistration({
                            archiverId: Config_1.config.ARCHIVER_IP,
                            timestamp: Date.now(),
                            validators: {
                                discovered: nodeList.length,
                                connected: nodeList.length,
                            },
                            state: 'REGISTERED',
                        });
                    }
                }
                catch (err) {
                    Logger.mainLogger.error('Error while joining network:');
                    Logger.mainLogger.error(err);
                    Logger.mainLogger.error(err.stack);
                    Logger.mainLogger.debug(`Trying to join again in ${cycleDuration} seconds...`);
                    const nodeList = NodeList.getActiveList();
                    archiverLogging_1.ArchiverLogging.logArchiverRegistration({
                        archiverId: Config_1.config.ARCHIVER_IP,
                        timestamp: Date.now(),
                        validators: {
                            discovered: nodeList.length,
                            connected: 0,
                        },
                        state: 'ERROR',
                    });
                    await Utils.sleep(cycleDuration);
                }
                firstTime = false;
            } while (!isJoined);
            /**
             * [NOTE] [AS] There's a possibility that we could get stuck in this loop
             * if the joinRequest was sent in the wrong cycle quarter (Q2, Q3, or Q4).
             *
             * Since we've dealt with this problem in shardus-global-server, it might be
             * good to refactor this code to do what shardus-global-server does to join
             * the network.
             */
            Logger.mainLogger.debug('We have successfully joined the network');
            await startServer();
            await Data.subscribeNodeForDataTransfer();
        }
        else {
            await startServer();
        }
    }
    else {
        try {
            Logger.mainLogger.debug('We are not first archiver. Syncing and starting archive-server');
            syncAndStartServer();
        }
        catch (err) {
            Logger.mainLogger.error('Error syncing and starting archive-server', err);
        }
    }
    setTimeout(() => {
        (0, transactionVerification_1.scheduleMultiSigKeysSyncFromNetConfig)();
    }, 60 * 1000); // Start after 60 seconds
    // Create the failed buckets directory
    (0, Utils_1.createDirectories)(Config_1.config.failedBucketsDir);
    // Initialize checkpoint V2 system if enabled and checkpoint updates and storage are allowed
    if (Config_1.config.checkpoint.bucketConfig.allowCheckpointUpdates) {
        Logger.mainLogger.info('Initializing checkpoint system...');
        // Start the update loop
        updateCheckpoints();
        // Sync missing checkpoints on startup if enabled
        if (Config_1.config.checkpoint.syncOnStartup) {
            setTimeout(async () => {
                try {
                    await (0, CheckpointV2_1.syncMissingCheckpoints)(Config_1.config.checkpoint.maxCyclesToSync);
                }
                catch (error) {
                    Logger.mainLogger.error('Error syncing missing checkpoints on startup:', error);
                }
            }, Config_1.config.checkpoint.syncInterval); // Wait 10 seconds after startup to begin syncing
        }
    }
    async function updateCheckpoints() {
        const startTime = Date.now();
        try {
            if (CycleData_1.cycleCheckpointManager && ReceiptData_1.receiptCheckpointManager && OriginalTxsData_1.originalTxCheckpointManager) {
                await Promise.allSettled([
                    CycleData_1.cycleCheckpointManager.update(),
                    ReceiptData_1.receiptCheckpointManager.update(),
                    OriginalTxsData_1.originalTxCheckpointManager.update(),
                ]);
            }
        }
        catch (error) {
            Logger.mainLogger.error('Error updating checkpoints:', error);
        }
        const elapsedTime = Date.now() - startTime;
        const nextExecutionDelay = Math.max(0, Config_1.config.checkpoint.updateInterval - elapsedTime);
        setTimeout(updateCheckpoints, nextExecutionDelay);
    }
}
function initProfiler(server) {
    const memoryReporter = new memoryReporting_1.default(server);
    (0, memoryReporting_1.setMemoryReportingInstance)(memoryReporter);
    const nestedCounter = new nestedCounters_1.default(server);
    (0, nestedCounters_1.setNestedCountersInstance)(nestedCounter);
    const profiler = new profiler_1.default(server);
    (0, profiler_1.setProfilerInstance)(profiler);
    const statistics = new statistics_1.default(logDir, Config_1.config.STATISTICS, {
        counters: [],
        watchers: {},
        timers: [],
        manualStats: ['cpuPercent'],
    }, {});
    statistics.startSnapshots();
    statistics.on('snapshot', memoryReporting_1.memoryReportingInstance.updateCpuPercent);
    // ========== ENDPOINTS ==========
    memoryReporter.registerEndpoints();
    nestedCounter.registerEndpoints();
    profiler.registerEndpoints();
}
/** Asynchronous function to synchronize and start the server. */
async function syncAndStartServer() {
    try {
        // Set the syncing flag to true to know we are patching the data
        State.setSyncing(true);
        // Get the cycle duration
        const cycleDuration = await Data.getCycleDuration();
        const oldestFailedCheckpointStatus = await (0, checkpointStatus_1.getOldestPendingOrFailedCheckpointStatus)();
        const firstUnifiedCheckpointCycle = oldestFailedCheckpointStatus?.cycle || 0;
        // Retrieve database state
        const lastStoredReceiptCount = await ReceiptDB.queryReceiptCount();
        const lastStoredCycleCount = await CycleDB.queryCyleCount();
        const lastStoredCycleInfo = (await CycleDB.queryLatestCycleRecords(1))[0];
        // Get the latest cycle from archivers to know how far we need to sync
        const latestNetworkCycle = await Cycles.getNewestCycleFromArchivers();
        State.setLastCycleToSync(latestNetworkCycle?.counter || oldestFailedCheckpointStatus?.cycle || 0);
        // Sync cycle data if checkpoint updates are allowed
        if (Config_1.config.checkpoint.bucketConfig.allowCheckpointUpdates) {
            await syncCycleDataWithCheckpoints(firstUnifiedCheckpointCycle, latestNetworkCycle);
        }
        else {
            await validateAndSyncCycleData(lastStoredCycleCount, lastStoredCycleInfo);
        }
        // Join the network
        await joinNetwork(cycleDuration);
        // Synchronize Genesis accounts and transactions from the network archivers
        await Data.syncGenesisAccountsFromArchiver();
        await Data.syncGenesisTransactionsFromArchiver();
        // Sync cycle and node list information
        if (Config_1.config.useSyncV2 === true) {
            await Data.syncCyclesAndNodeListV2(State.activeArchivers, lastStoredCycleCount);
        }
        else {
            await Data.syncCyclesAndNodeList(lastStoredCycleCount);
        }
        // Handle receipts and transaction data based on configuration
        if (Config_1.config.experimentalSnapshot) {
            await (0, GlobalAccount_1.syncGlobalAccount)();
            if (Config_1.config.checkpoint.bucketConfig.allowCheckpointUpdates) {
                await handleReceiptSyncWithCheckpoints(lastStoredReceiptCount, firstUnifiedCheckpointCycle, latestNetworkCycle);
            }
            else {
                await handleTraditionalReceiptSync(lastStoredReceiptCount, lastStoredCycleCount);
            }
        }
        else {
            // Sync all state metadata until no older data is fetched from other archivers
            await (0, StateMetaData_1.syncStateMetaData)(State.activeArchivers);
            // Wait for one cycle before sending data request if experimentalSnapshot is not enabled
            await Utils.sleep(cycleDuration * 1000);
            // Start the server
            await startServer();
        }
        // Sync any missing data that may have been created during startup
        const latestCycle = await Cycles.getNewestCycleFromArchivers();
        const beforeCycle = Cycles.getCurrentCycleCounter();
        await Data.syncCyclesAndTxsDataBetweenCycles(beforeCycle - 1, latestCycle.counter + 1);
        // Send active message if configured
        if (Config_1.config.sendActiveMessage) {
            await sendActiveMessage();
        }
        else {
            State.setSyncing(false);
            await Data.subscribeNodeForDataTransfer();
        }
        // Schedule multi-signature keys sync
        (0, transactionVerification_1.scheduleMultiSigKeysSyncFromNetConfig)();
    }
    catch (error) {
        Logger.mainLogger.error('Error in syncAndStartServer:', error);
        throw error;
    }
}
// Helper functions for syncAndStartServer
async function syncCycleDataWithCheckpoints(firstUnifiedCheckpointCycle, latestNetworkCycle) {
    const response = await Data.getTotalDataFromArchivers();
    const { totalCycles } = response;
    if (firstUnifiedCheckpointCycle > totalCycles) {
        Logger.mainLogger.info('The existing db has more data than the network data! Proceeding with joining the network!');
        return;
    }
    const startCycle = firstUnifiedCheckpointCycle;
    let endCycle = Math.max(totalCycles, latestNetworkCycle?.counter || 0);
    const BATCH_SIZE = Config_1.config.checkpoint.batchSize;
    let currentStart = startCycle;
    let currentEnd = Math.min(currentStart + BATCH_SIZE, endCycle);
    Logger.mainLogger.info(`Need to patch cycles from ${currentStart} to ${currentEnd}...`);
}
async function validateAndSyncCycleData(lastStoredCycleCount, lastStoredCycleInfo) {
    const response = await Data.getTotalDataFromArchivers();
    if (!response ||
        response.totalCycles < 0 ||
        response.totalAccounts < 0 ||
        response.totalTransactions < 0 ||
        response.totalReceipts < 0) {
        throw Error(`Can't fetch total data from archivers`);
    }
    const { totalCycles, totalReceipts } = response;
    const lastStoredReceiptCount = await ReceiptDB.queryReceiptCount();
    if (lastStoredReceiptCount > totalReceipts || lastStoredCycleCount > totalCycles) {
        throw Error('The existing db has more data than the network data! Clear the DB and start the server again!');
    }
    if (lastStoredCycleCount > 0) {
        Logger.mainLogger.debug('Validating old cycles data!');
        const cycleResult = await Data.compareWithOldCyclesData(lastStoredCycleInfo.counter);
        if (!cycleResult.success) {
            throw Error('The last saved 10 cycles data does not match with the archiver data! Clear the DB and start the server again!');
        }
    }
}
async function joinNetwork(cycleDuration) {
    let isJoined = false;
    let firstTime = true;
    do {
        try {
            const randomArchiver = State.getRandomArchiver();
            const nodeList = await NodeList.getActiveNodeListFromArchiver(randomArchiver);
            if (nodeList.length === 0)
                continue;
            isJoined = await Data.joinNetwork(nodeList, firstTime);
        }
        catch (err) {
            Logger.mainLogger.error('Error while joining network:', err);
            await Utils.sleep(cycleDuration * 1000);
        }
        firstTime = false;
    } while (!isJoined);
    /**
     * [NOTE] [AS] There's a possibility that we could get stuck in this loop
     * if the joinRequest was sent in the wrong cycle quarter (Q2, Q3, or Q4).
     *
     * Since we've dealt with this problem in shardus-global-server, it might be
     * good to refactor this code to do what shardus-global-server does to join
     * the network.
     */
    Logger.mainLogger.debug('We have successfully joined the network');
}
async function handleReceiptSyncWithCheckpoints(lastStoredReceiptCount, firstUnifiedCheckpointCycle, latestNetworkCycle) {
    Logger.mainLogger.info('Using checkpoint V2 for data synchronization');
    // Find the last valid cycle for receipts
    let lastStoredReceiptCycle = Math.max(firstUnifiedCheckpointCycle - 1, 0);
    // Sync receipts based on what's stored
    if (lastStoredReceiptCount === 0) {
        await Data.syncReceipts();
    }
    else {
        await Data.syncReceiptsByCycle(lastStoredReceiptCycle);
    }
    // Verify database integrity after syncing
    const updatedReceiptCount = await ReceiptDB.queryReceiptCount();
    const updatedCycleCount = await CycleDB.queryCyleCount();
    const updatedCycleInfo = (await CycleDB.queryLatestCycleRecords(1))[0];
    if (updatedCycleCount - 1 !== updatedCycleInfo.counter) {
        throw Error(`The archiver has ${updatedCycleCount} cycles but the latest stored cycle is ${updatedCycleInfo.counter}`);
    }
    await Data.syncCyclesAndTxsData(updatedCycleCount, updatedReceiptCount);
    // Start the server first, then sync missing data in the background
    await startServer();
    // Schedule background sync
    setTimeout(async () => {
        try {
            const { syncMissingCheckpoints } = await Promise.resolve().then(() => require('./checkpoint/CheckpointV2'));
            await syncMissingCheckpoints(Config_1.config.checkpoint.maxCyclesToSync);
        }
        catch (error) {
            Logger.mainLogger.error('Error syncing missing checkpoints after server start:', error);
        }
    }, Config_1.config.checkpoint.syncInterval);
}
async function handleTraditionalReceiptSync(lastStoredReceiptCount, lastStoredCycleCount) {
    if (lastStoredReceiptCount === 0) {
        await Data.syncReceipts();
    }
    else {
        const lastStoredReceiptInfo = await ReceiptDB.queryLatestReceipts(1);
        const lastStoredReceiptCycle = lastStoredReceiptInfo?.[0]?.cycle || 0;
        Logger.mainLogger.debug('lastStoredReceiptCycle', lastStoredReceiptCycle);
        await Data.syncReceiptsByCycle(lastStoredReceiptCycle);
    }
    // Verify database integrity
    const updatedReceiptCount = await ReceiptDB.queryReceiptCount();
    const updatedCycleCount = await CycleDB.queryCyleCount();
    const updatedCycleInfo = (await CycleDB.queryLatestCycleRecords(1))[0];
    if (updatedCycleCount - 1 !== updatedCycleInfo.counter) {
        throw Error(`The archiver has ${updatedCycleCount} cycles but the latest stored cycle is ${updatedCycleInfo.counter}`);
    }
    await Data.syncCyclesAndTxsData(updatedCycleCount, updatedReceiptCount);
    await startServer();
}
async function sendActiveMessage() {
    let isActive = false;
    while (!isActive) {
        await Data.sendActiveRequest();
        // TODO not used for now
        // isActive = await Data.checkActiveStatus()
        // Set as true for now, This needs to be removed after the active record for the archiver is added on the validator side
        isActive = true;
    }
    await Data.subscribeNodeForDataTransfer();
    State.setSyncing(false);
}
// Define all endpoints, all requests, and start REST server
async function startServer() {
    const server = (0, fastify_1.default)({
        logger: false,
    });
    await server.register(cors_1.default);
    await server.register(rate_limit_1.default, {
        global: true,
        max: Config_1.config.RATE_LIMIT,
        timeWindow: 10,
        allowList: ['127.0.0.1', '0.0.0.0'], // Excludes local IPs from rate limits
    });
    await server.register(healthCheck_1.healthCheckRouter);
    server.addContentTypeParser('application/json', { parseAs: 'string' }, (req, body, done) => {
        try {
            const jsonString = typeof body === 'string' ? body : body.toString('utf8');
            done(null, lib_types_1.Utils.safeJsonParse(jsonString));
        }
        catch (err) {
            err.statusCode = 400;
            done(err, undefined);
        }
    });
    server.setReplySerializer((payload) => {
        return lib_types_1.Utils.safeStringify(payload);
    });
    initProfiler(server);
    // Initialize the data log writer
    if (Config_1.config.dataLogWrite)
        await (0, DataLogWriter_1.initDataLogWriter)();
    // Register API routes
    (0, API_1.registerRoutes)(server);
    // Start server and bind to port on all interfaces
    server.listen({
        port: Config_1.config.ARCHIVER_PORT,
        host: '0.0.0.0',
    }, (err) => {
        if (err) {
            server.log.error(err);
            process.exit(1);
        }
        Logger.mainLogger.info(`Worker ${process.pid}: Archive-server is listening on http://0.0.0.0:${Config_1.config.ARCHIVER_PORT}`);
        State.setActive();
        Collector.scheduleMissingTxsDataQuery();
        // setupWorkerProcesses(cluster)
    });
}
// Add this before starting the server
try {
    (0, tickets_1.initializeTickets)();
}
catch (err) {
    console.error('Failed to initialize tickets. Server startup aborted:', err);
    process.exit(1);
}
start();
//# sourceMappingURL=server.js.map