"use strict";
/**
 * SyncV2 a p2p module that contains all of the functionality for the new
 * Node List Sync v2.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.syncTxList = exports.syncV2 = void 0;
const neverthrow_1 = require("neverthrow");
const queries_1 = require("./queries");
const State_1 = require("../State");
const NodeList_1 = require("../NodeList");
const NodeList = require("../NodeList");
const verify_1 = require("./verify");
const Logger = require("../Logger");
const ServiceQueue = require("../ServiceQueue");
const archiverLogging_1 = require("../profiler/archiverLogging");
const Config_1 = require("../Config");
/**
 * Given a list of archivers, queries each one until one returns an active node list.
 *
 * The endpoint queried does not return a *full* list of nodes. It's a partial
 * list that will be enough to use in robust queries.
 */
async function getActiveListFromSomeArchiver(archivers, operationId) {
    const startTime = Date.now();
    archiverLogging_1.ArchiverLogging.logDataSync({
        sourceArchiver: archivers[0].ip,
        targetArchiver: Config_1.config.ARCHIVER_IP,
        cycle: 0,
        dataType: 'ARCHIVER_LIST',
        dataHash: '',
        status: 'STARTED',
        operationId,
        metrics: {
            duration: 0,
            dataSize: 0,
        },
    });
    for (const archiver of archivers) {
        try {
            const nodeList = await (0, NodeList_1.getActiveNodeListFromArchiver)(archiver);
            if (nodeList) {
                archiverLogging_1.ArchiverLogging.logDataSync({
                    sourceArchiver: archiver.ip,
                    targetArchiver: Config_1.config.ARCHIVER_IP,
                    cycle: 0,
                    dataType: 'ARCHIVER_LIST',
                    dataHash: '',
                    status: 'IN_PROGRESS',
                    operationId,
                    metrics: {
                        duration: Date.now() - startTime,
                        dataSize: JSON.stringify(nodeList).length,
                    },
                });
                return nodeList;
            }
        }
        catch (e) {
            console.warn(`failed to get active node list from archiver ${archiver.ip}:${archiver.port}: ${e}`);
            archiverLogging_1.ArchiverLogging.logDataSync({
                sourceArchiver: archiver.ip,
                targetArchiver: Config_1.config.ARCHIVER_IP,
                cycle: 0,
                dataType: 'ARCHIVER_LIST',
                dataHash: '',
                status: 'ERROR',
                operationId,
                metrics: {
                    duration: Date.now() - startTime,
                    dataSize: 0,
                },
                error: e.message,
            });
            continue;
        }
    }
    // all archivers have failed at this point
    throw new Error('no archiver could return an active node list');
}
/**
 * Synchronizes the NodeList and gets the latest CycleRecord from other validators.
 */
function syncV2(activeArchivers) {
    const startTime = Date.now();
    const operationId = archiverLogging_1.ArchiverLogging.generateOperationId();
    return neverthrow_1.ResultAsync.fromPromise(getActiveListFromSomeArchiver(activeArchivers, operationId), (e) => e).andThen((nodeList) => syncValidatorList(nodeList, operationId).andThen(([validatorList, validatorListHash]) => syncArchiverList(nodeList, operationId).andThen(([archiverList, archiverListHash]) => syncStandbyNodeList(nodeList, operationId).andThen(([standbyList, standbyListHash]) => syncTxList(nodeList, operationId).andThen((txList) => syncLatestCycleRecordAndMarker(nodeList, operationId).andThen(([cycle, cycleMarker]) => {
        Logger.mainLogger.debug('syncV2: validatorList', validatorList);
        archiverLogging_1.ArchiverLogging.logDataSync({
            sourceArchiver: nodeList[0].ip,
            targetArchiver: Config_1.config.ARCHIVER_IP,
            cycle: cycle.counter,
            dataType: 'CYCLE_RECORD',
            dataHash: cycleMarker,
            status: 'COMPLETE',
            operationId,
            metrics: {
                duration: Date.now() - startTime,
                dataSize: JSON.stringify({
                    validatorList,
                    archiverList,
                    standbyList,
                    txList,
                    cycle,
                }).length,
            },
        });
        // additional checks to make sure the list hashes in the cycle
        // matches the hash for the validator list retrieved earlier
        if (cycle.nodeListHash !== validatorListHash) {
            archiverLogging_1.ArchiverLogging.logDataSync({
                sourceArchiver: nodeList[0].ip,
                targetArchiver: Config_1.config.ARCHIVER_IP,
                cycle: cycle.counter,
                dataType: 'VALIDATOR_LIST',
                dataHash: validatorListHash,
                status: 'ERROR',
                operationId,
                metrics: {
                    duration: Date.now() - startTime,
                    dataSize: JSON.stringify(validatorList).length,
                },
                error: `validator list hash from received cycle (${cycle.nodeListHash}) does not match the hash received from robust query (${validatorListHash})`,
            });
            return (0, neverthrow_1.errAsync)(new Error(`validator list hash from received cycle (${cycle.nodeListHash}) does not match the hash received from robust query (${validatorListHash})`));
        }
        if (cycle.standbyNodeListHash !== standbyListHash) {
            archiverLogging_1.ArchiverLogging.logDataSync({
                sourceArchiver: nodeList[0].ip,
                targetArchiver: Config_1.config.ARCHIVER_IP,
                cycle: cycle.counter,
                dataType: 'STANDBY_LIST',
                dataHash: standbyListHash,
                status: 'ERROR',
                operationId,
                metrics: {
                    duration: Date.now() - startTime,
                    dataSize: JSON.stringify(standbyList).length,
                },
                error: `standby list hash from received cycle (${cycle.standbyNodeListHash}) does not match the hash received from robust query (${standbyListHash})`,
            });
            return (0, neverthrow_1.errAsync)(new Error(`standby list hash from received cycle (${cycle.nodeListHash}) does not match the hash received from robust query (${validatorListHash})`));
        }
        if (cycle.archiverListHash !== archiverListHash) {
            archiverLogging_1.ArchiverLogging.logDataSync({
                sourceArchiver: nodeList[0].ip,
                targetArchiver: Config_1.config.ARCHIVER_IP,
                cycle: cycle.counter,
                dataType: 'ARCHIVER_LIST',
                dataHash: archiverListHash,
                status: 'ERROR',
                operationId,
                metrics: {
                    duration: Date.now() - startTime,
                    dataSize: JSON.stringify(archiverList).length,
                },
                error: `archiver list hash from received cycle (${cycle.archiverListHash}) does not match the hash received from robust query (${archiverListHash})`,
            });
            return (0, neverthrow_1.errAsync)(new Error(`archiver list hash from received cycle (${cycle.archiverListHash}) does not match the hash received from robust query (${archiverListHash})`));
        }
        // validatorList and standbyList need to be transformed into a ConsensusNodeInfo[]
        const syncingNodeList = [];
        const activeNodeList = [];
        for (const node of validatorList) {
            if (node.status === 'selected' || node.status === 'syncing' || node.status === 'ready') {
                syncingNodeList.push({
                    publicKey: node.publicKey,
                    ip: node.externalIp,
                    port: node.externalPort,
                    id: node.id,
                });
            }
            else if (node.status === 'active') {
                activeNodeList.push({
                    publicKey: node.publicKey,
                    ip: node.externalIp,
                    port: node.externalPort,
                    id: node.id,
                });
            }
        }
        const standbyNodeList = standbyList.map((joinRequest) => ({
            publicKey: joinRequest.nodeInfo.publicKey,
            ip: joinRequest.nodeInfo.externalIp,
            port: joinRequest.nodeInfo.externalPort,
        }));
        NodeList.addNodes(NodeList.NodeStatus.SYNCING, syncingNodeList);
        NodeList.addNodes(NodeList.NodeStatus.ACTIVE, activeNodeList);
        NodeList.addStandbyNodes(standbyNodeList);
        // add txList
        ServiceQueue.setTxList(txList);
        // reset the active archivers list with the new list
        (0, State_1.resetActiveArchivers)(archiverList);
        // return a cycle that we'll store in the database
        return (0, neverthrow_1.okAsync)({
            ...cycle,
            marker: cycleMarker,
        });
    }))))));
}
exports.syncV2 = syncV2;
/**
 * This function synchronizes a validator list from `activeNodes`.
 *
 * @param {P2PTypes.SyncTypes.ActiveNode[]} activeNodes - An array of active nodes to be queried.
 * The function first performs a robust query for the latest node list hash.
 * After obtaining the hash, it retrieves the full node list from one of the winning nodes.
 *
 * @returns {ResultAsync<P2PTypes.NodeListTypes.Node[], Error>} - A ResultAsync object. On success, it will contain
 * an array of Node objects, and on error, it will contain an Error object. The function is asynchronous
 * and can be awaited.
 */
function syncValidatorList(activeNodes, operationId) {
    const startTime = Date.now();
    archiverLogging_1.ArchiverLogging.logDataSync({
        sourceArchiver: activeNodes[0].ip,
        targetArchiver: Config_1.config.ARCHIVER_IP,
        cycle: 0,
        dataType: 'VALIDATOR_LIST',
        dataHash: '',
        status: 'STARTED',
        operationId,
        metrics: {
            duration: 0,
            dataSize: 0,
        },
    });
    return (0, queries_1.robustQueryForValidatorListHash)(activeNodes).andThen(({ value, winningNodes }) => (0, queries_1.getValidatorListFromNode)(winningNodes[0], value.nodeListHash).andThen((validatorList) => (0, verify_1.verifyValidatorList)(validatorList, value.nodeListHash).map(() => {
        archiverLogging_1.ArchiverLogging.logDataSync({
            sourceArchiver: activeNodes[0].ip,
            targetArchiver: Config_1.config.ARCHIVER_IP,
            cycle: 0,
            dataType: 'VALIDATOR_LIST',
            dataHash: value.nodeListHash,
            status: 'IN_PROGRESS',
            operationId,
            metrics: {
                duration: Date.now() - startTime,
                dataSize: JSON.stringify(validatorList).length,
            },
        });
        return [validatorList, value.nodeListHash];
    })));
}
/**
 * This function synchronizes a standby node list from `activeNodes`.
 *
 * @param {P2PTypes.SyncTypes.ActiveNode[]} activeNodes - An array of active nodes to be queried.
 * The function first performs a robust query for the latest node list hash.
 * After obtaining the hash, it retrieves the full node list from one of the winning nodes.
 *
 * @returns {ResultAsync<P2PTypes.NodeListTypes.Node[], Error>} - A ResultAsync object. On success, it will contain
 * an array of Node objects, and on error, it will contain an Error object. The function is asynchronous
 * and can be awaited.
 */
function syncStandbyNodeList(activeNodes, operationId) {
    const startTime = Date.now();
    archiverLogging_1.ArchiverLogging.logDataSync({
        sourceArchiver: activeNodes[0].ip,
        targetArchiver: Config_1.config.ARCHIVER_IP,
        cycle: 0,
        dataType: 'STANDBY_LIST',
        dataHash: '',
        status: 'STARTED',
        operationId,
        metrics: {
            duration: 0,
            dataSize: 0,
        },
    });
    return (0, queries_1.robustQueryForStandbyNodeListHash)(activeNodes).andThen(({ value, winningNodes }) => (0, queries_1.getStandbyNodeListFromNode)(winningNodes[0], value.standbyNodeListHash).andThen((standbyList) => {
        archiverLogging_1.ArchiverLogging.logDataSync({
            sourceArchiver: activeNodes[0].ip,
            targetArchiver: Config_1.config.ARCHIVER_IP,
            cycle: 0,
            dataType: 'STANDBY_LIST',
            dataHash: value.standbyNodeListHash,
            status: 'IN_PROGRESS',
            operationId,
            metrics: {
                duration: Date.now() - startTime,
                dataSize: JSON.stringify(standbyList).length,
            },
        });
        return (0, neverthrow_1.okAsync)([standbyList, value.standbyNodeListHash]);
    }));
}
function syncTxList(activeNodes, operationId) {
    const startTime = Date.now();
    archiverLogging_1.ArchiverLogging.logDataSync({
        sourceArchiver: activeNodes[0].ip,
        targetArchiver: Config_1.config.ARCHIVER_IP,
        cycle: 0,
        dataType: 'TX_LIST',
        dataHash: '',
        status: 'STARTED',
        operationId,
        metrics: {
            duration: 0,
            dataSize: 0,
        },
    });
    return (0, queries_1.robustQueryForTxListHash)(activeNodes).andThen(({ value, winningNodes }) => (0, queries_1.getTxListFromNode)(winningNodes[0], value.txListHash).andThen((txList) => (0, verify_1.verifyTxList)(txList, value.txListHash).map(() => {
        archiverLogging_1.ArchiverLogging.logDataSync({
            sourceArchiver: activeNodes[0].ip,
            targetArchiver: Config_1.config.ARCHIVER_IP,
            cycle: 0,
            dataType: 'TX_LIST',
            dataHash: value.txListHash,
            status: 'IN_PROGRESS',
            operationId,
            metrics: {
                duration: Date.now() - startTime,
                dataSize: JSON.stringify(txList).length,
            },
        });
        return txList;
    })));
}
exports.syncTxList = syncTxList;
/**
 * Synchronizes the latest cycle record from a list of active nodes.
 *
 * @param {P2PTypes.SyncTypes.ActiveNode[]} activeNodes - An array of active nodes to be queried.
 * The function first performs a robust query for the latest cycle record hash.
 * After obtaining the hash, it retrieves the current cycle data from one of the winning nodes.
 *
 * @returns {ResultAsync<P2PTypes.CycleCreatorTypes.CycleRecord, Error>} - A ResultAsync object.
 * On success, it will contain a CycleRecord object, and on error, it will contain an Error object.
 * The function is asynchronous and can be awaited.
 */
function syncLatestCycleRecordAndMarker(activeNodes, operationId) {
    const startTime = Date.now();
    archiverLogging_1.ArchiverLogging.logDataSync({
        sourceArchiver: activeNodes[0].ip,
        targetArchiver: Config_1.config.ARCHIVER_IP,
        cycle: 0,
        dataType: 'CYCLE_RECORD',
        dataHash: '',
        status: 'STARTED',
        operationId,
        metrics: {
            duration: 0,
            dataSize: 0,
        },
    });
    return (0, queries_1.robustQueryForCycleRecordHash)(activeNodes).andThen(({ value, winningNodes }) => (0, queries_1.getCurrentCycleDataFromNode)(winningNodes[0], value.currentCycleHash).andThen((cycle) => (0, verify_1.verifyCycleRecord)(cycle, value.currentCycleHash).map(() => {
        archiverLogging_1.ArchiverLogging.logDataSync({
            sourceArchiver: activeNodes[0].ip,
            targetArchiver: Config_1.config.ARCHIVER_IP,
            cycle: cycle.counter,
            dataType: 'CYCLE_RECORD',
            dataHash: value.currentCycleHash,
            status: 'IN_PROGRESS',
            operationId,
            metrics: {
                duration: Date.now() - startTime,
                dataSize: JSON.stringify(cycle).length,
            },
        });
        return [cycle, value.currentCycleHash];
    })));
}
/**
 * This function queries for an archiver list from other active nodes.
 *
 * @param {P2P.SyncTypes.ActiveNode[]} activeNodes - An array of active nodes to be queried.
 * The function first performs a robust query for the latest archiver list hash.
 * Then, it requests a full list from one of the winning nodes using the hash
 * retrieved. The node receiving the request may or may not have the list whose
 * hash matches the one requested.
 *
 * @returns {ResultAsync<[P2P.ArchiversTypes.JoinedArchiver[], hexstring], Error>} - A ResultAsync object. On success, it will contain an array of
 * JoinedArchiver objects and the archiver list hash, and on error, it will contain an Error object. The function is asynchronous and can be awaited.
 */
function syncArchiverList(activeNodes, operationId) {
    const startTime = Date.now();
    archiverLogging_1.ArchiverLogging.logDataSync({
        sourceArchiver: activeNodes[0].ip,
        targetArchiver: Config_1.config.ARCHIVER_IP,
        cycle: 0,
        dataType: 'ARCHIVER_LIST',
        dataHash: '',
        status: 'STARTED',
        operationId,
        metrics: {
            duration: 0,
            dataSize: 0,
        },
    });
    // run a robust query for the lastest archiver list hash
    return (0, queries_1.robustQueryForArchiverListHash)(activeNodes).andThen(({ value, winningNodes }) => 
    // get full archiver list from one of the winning nodes
    (0, queries_1.getArchiverListFromNode)(winningNodes[0], value.archiverListHash).andThen((archiverList) => 
    // verify a hash of the retrieved archiver list matches the hash from before.
    // if it does, return the archiver list
    (0, verify_1.verifyArchiverList)(archiverList, value.archiverListHash).map(() => {
        archiverLogging_1.ArchiverLogging.logDataSync({
            sourceArchiver: activeNodes[0].ip,
            targetArchiver: Config_1.config.ARCHIVER_IP,
            cycle: 0,
            dataType: 'ARCHIVER_LIST',
            dataHash: value.archiverListHash,
            status: 'IN_PROGRESS',
            operationId,
            metrics: {
                duration: Date.now() - startTime,
                dataSize: JSON.stringify(archiverList).length,
            },
        });
        return [archiverList, value.archiverListHash];
    })));
}
//# sourceMappingURL=index.js.map